"""Some shared core functionality in the package.

This contains mostly wrappers around `xarray` objects, and some very basic functions.
A few of these are publicly available in the main package namespace. They should be
accessed from there if used externally, but from here if used internally.


Classes and functions only exposed here
---------------------------------------
.. autosummary::
    :toctree: generated

    time_to_np
    time_to_datetime
    time_frame_settings
    TimeWindow
    xrwrap
    DataArrayWrap
    DatasetWrap
    Roller
    TimeDataRoller

Classes and functions exposed in the main package namespace
-----------------------------------------------------------
.. autosummary::

    uwacan.TimeData
    uwacan.FrequencyData
    uwacan.TimeFrequencyData
    uwacan.Transit
    uwacan.dB

"""

import numpy as np
import xarray as xr
import whenever

__all__ = [
    "TimeWindow",
    "dB",
    "TimeData",
    "FrequencyData",
    "TimeFrequencyData",
    "Transit",
]


def time_to_np(input, **kwargs):
    """Convert a time to `numpy.datetime64`."""
    if isinstance(input, np.datetime64):
        return input
    if not isinstance(input, whenever.Instant):
        input = time_to_datetime(input, **kwargs)
    return np.datetime64(input.timestamp_nanos(), "ns")


def time_to_datetime(input, fmt="RFC 3339", tz="UTC"):
    """Convert datetimes to the same internal format.

    This function takes a few types of input and tries to convert
    the input to a `whenever.Instant`.
    - Any datetime-like input will be converted directly.
    - np.datetime64 and Unix timestamps are treated similarly.
    - Strings are parsed with ``fmt`` if given, otherwise a few different common formats are tried.

    Parameters
    ----------
    input : datetime-like, string, or numeric.
        The input data specifying the time.
    fmt : string, optional
        This can be one of the standard formats ``"RFC 3339"``, ``"RFC 2822"``, or ``"ISO 8601"``, handled by `whenever <https://whenever.readthedocs.io/en/latest/overview.html#formatting-and-parsing>`_.
        Alternatively, a custom parsing specifier can be supplied, using `strptime format codes <https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes>`_.
    tz : string, default "UTC"
        The assumed timezone for the input, if it does not contain one.
        Unix timestamps have no timezone, and np.datetime64 only supports UTC.

    Returns
    -------
    time : whenever.Instant
        The converted time.
    """
    if isinstance(input, str):
        if input == "now":
            return whenever.Instant.now()
        if fmt == "RFC 3339":
            return whenever.OffsetDateTime.parse_rfc3339(input).instant()
        if fmt == "RFC 2822":
            return whenever.OffsetDateTime.parse_rfc2822(input).instant()
        if "ISO" in fmt:
            return whenever.OffsetDateTime.parse_common_iso(input).instant()
        if "z" in fmt:
            return whenever.OffsetDateTime.strptime(input, fmt).instant()
        if tz == "UTC":
            return whenever.LocalDateTime.strptime(input, fmt).assume_utc()
        if "/" in tz:
            return whenever.LocalDateTime.strptime(input, fmt).assume_tz(tz).instant()
        if ":" in tz:
            if tz[0] == "+":
                sign = 1
                tz = tz[1:]
            elif tz[0] == "-":
                sign = -1
                tz = tz[1:]
            else:
                sign = 1
            hours, minutes = tz.split(":")
            tz = sign * whenever.TimeDelta(hours=int(hours), minutes=int(minutes))
            return whenever.LocalDateTime.strptime(input, fmt).assume_fixed_offset(tz).instant()

    if isinstance(input, whenever.Instant):
        return input
    if isinstance(input, (whenever.OffsetDateTime, whenever.ZonedDateTime)):
        return input.instant()
    if isinstance(input, whenever.LocalDateTime):
        return input.assume_utc()

    if hasattr(input, "timestamp"):
        if callable(input.timestamp):
            input = input.timestamp()
        else:
            input = input.timestamp
    if isinstance(input, xr.DataArray):
        input = input.data
    if isinstance(input, np.datetime64) or (
        isinstance(input, np.ndarray) and np.issubdtype(input.dtype, np.datetime64)
    ):
        if tz != "UTC":
            raise ValueError("Numpy datetime64 values should always be stored in UTC")
        input = float(input.astype("timedelta64") / np.timedelta64(1, "s"))
    return whenever.Instant.from_timestamp(input)


class TimeWindow:
    """Describes a start and stop point in time.

    Give two and only two of the four basic parameters.
    Less than two will not fully define a window, while
    more than two will overdetermine the window.

    Parameters
    ----------
    start : time_like
        A window that starts at this time
    stop : time_like
        A window stat stops at this time
    center : time_like
        A window centered at this time
    duration : float
        A window with this duration, in seconds
    extend : float
        Extend the window defined by two of the four above
        with this much at each end, in seconds.
    """

    def __init__(self, start=None, stop=None, center=None, duration=None, extend=None):
        if start is not None:
            start = time_to_datetime(start)
        if stop is not None:
            stop = time_to_datetime(stop)
        if center is not None:
            center = time_to_datetime(center)

        if None not in (start, stop):
            _start = start
            _stop = stop
            start = stop = None
        elif None not in (center, duration):
            _start = center.subtract(seconds=duration / 2)
            _stop = center.add(seconds=duration / 2)
            center = duration = None
        elif None not in (start, duration):
            _start = start
            _stop = start.add(seconds=duration)
            start = duration = None
        elif None not in (stop, duration):
            _stop = stop
            _start = stop.subtract(seconds=duration)
            stop = duration = None
        elif None not in (start, center):
            _start = start
            _stop = start + (center - start) * 2
            start = center = None
        elif None not in (stop, center):
            _stop = stop
            _start = stop - (stop - center) * 2
            stop = center = None
        else:
            raise TypeError("Needs two of the input arguments to determine time window.")

        if (start, stop, center, duration) != (None, None, None, None):
            raise TypeError("Cannot input more than two input arguments to a time window!")

        if extend is not None:
            _start = _start.subtract(seconds=extend)
            _stop = _stop.add(seconds=extend)

        self._start = _start
        self._stop = _stop

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        """Select a smaller window of time.

        Parameters
        ----------
        time : time_window_like
            An object that will be used to extract start and stop times.
        start : time_like
            A new window that starts at this time.
            Give ``True`` to use the start of the existing window.
        stop : time_like
            A new window stat stops at this time
            Give ``True`` to use the stop of the existing window.
        center : time_like
            A new window centered at this time
            Give ``True`` to use the center of the existing window.
        duration : float
            A new window with this duration, in seconds
        extend : float
            Extend the new window defined by two of the four above
            with this much at each end, in seconds.

        Notes
        -----
        This takes the same basic inputs as `TimeWindow`, defining a window
        with two out of four of ``start``, ``stop``, ``center``, and ``duration``.
        Additionally, one of ``start``, ``stop``, ``center`` can be given as ``True``
        instead of an actual time to use the times of the existing window.
        If only one of ``start`` and ``stop`` is given, the other one is filled from
        the existing window.

        If a single positional argument is given, it should be time_window_like,
        i.e., have a defined start and stop time, which will then be used.
        This can be one of `TimeWindow`, and `xarray.Dataset`.
        If it is a dataset, it must have a time attribute, and its minimum and maximum
        will be used as the start and stop for the new window.
        """
        if time is None:
            # Period specified with keyword arguments, convert to period.
            if (start, stop, center, duration).count(None) == 3:
                # Only one argument which has to be start or stop, fill the other from self.
                if start is not None:
                    window = type(self)(start=start, stop=self.stop, extend=extend)
                elif stop is not None:
                    window = type(self)(start=self.start, stop=stop, extend=extend)
                else:
                    raise TypeError("Cannot create subwindow from arguments")
            elif duration is not None and True in (start, stop, center):
                if start is True:
                    window = type(self)(start=self.start, duration=duration, extend=extend)
                elif stop is True:
                    window = type(self)(stop=self.stop, duration=duration, extend=extend)
                elif center is True:
                    window = type(self)(center=self.center, duration=duration, extend=extend)
                else:
                    raise TypeError("Cannot create subwindow from arguments")
            else:
                # The same types explicit arguments as the normal constructor
                window = type(self)(start=start, stop=stop, center=center, duration=duration, extend=extend)
        elif isinstance(time, type(self)):
            window = time
        elif isinstance(time, xr.Dataset):
            window = type(self)(start=time.time.min(), stop=time.time.max(), extend=extend)
        else:
            # It's not a period, so it should be a single datetime. Parse or convert, check validity.
            time = time_to_datetime(time)
            if time not in self:
                raise ValueError("Received time outside of contained window")
            return time

        if window not in self:
            raise ValueError("Requested subwindow is outside contained time window")
        return window

    def __repr__(self):
        return f"TimeWindow(start={self.start.format_rfc3339()}, stop={self.stop.format_rfc3339()})"

    @property
    def start(self):
        """The start of this window."""
        return self._start

    @property
    def stop(self):
        """The stop of this window."""
        return self._stop

    @property
    def center(self):
        """The center of this window."""
        return self.start.add(seconds=self.duration / 2)

    @property
    def duration(self):
        """The duration of this window, in seconds."""
        return (self.stop - self.start).in_seconds()

    def __contains__(self, other):
        if isinstance(other, type(self)):
            return other.start in self and other.stop in self
        return self.start <= time_to_datetime(other) <= self.stop


class xrwrap:
    """Wrapper around `xarray` objects.

    This base class exists to delegate work to our internal
    `xarray` objects.
    """

    @classmethod
    def from_dataset(cls, dataset, **kwargs):
        """Instantiate the class from a dataset.

        This classmethod is mainly used to choose the correct class
        to instantiate, depending on the data.
        """
        return cls(dataset, **kwargs)

    def __init__(self, data, attrs=None):
        if isinstance(data, xrwrap):
            data = data.data
        self._data = data
        if attrs:
            self._data.attrs.update(attrs)

    @property
    def attrs(self):
        """Attributes stored in the data."""
        return self.data.attrs

    @property
    def data(self):
        """The contained data."""
        return self._data

    def __array_wrap__(self, data, context=None, transfer_attributes=True):
        """Wrap output data in in a new object.

        This takes data from some processing and wraps it back into a
        suitable class. If no suitable class was found, the data is
        returned as is.
        """
        try:
            data = self.from_dataset(data)
        except NotImplementedError:
            return data
        if transfer_attributes:
            data.attrs.update(self.attrs)
        return data

    def sel(self, indexers=None, method=None, tolerance=None, drop=False, drop_allnan=True, **indexers_kwargs):
        """Select a subset of the data from the coordinate labels.

        The selection is easiest done with keywords, e.g. ``obj.sel(sensor="Colmar 1")``
        to select a specific sensor. For numerical coordinates, ``method="nearest"`` can
        be quite useful. Use a slice to select a range of values, e.g.,
        ``obj.sel(frequency=slice(10, 100))``.

        For more details, see `xarray.DataArray.sel` and `xarray.Dataset.sel`.
        """
        new = self.data.sel(indexers=indexers, method=method, tolerance=tolerance, drop=drop, **indexers_kwargs)
        if drop_allnan:
            new = new.where(~new.isnull(), drop=True)
        return self.__array_wrap__(new)

    def isel(self, indexers=None, drop=False, missing_dims="raise", drop_allnan=True, **indexers_kwargs):
        """Select a subset of the data from the coordinate indices.

        The selection is easiest done with keywords, e.g. ``obj.sel(sensor=0)``
        to select the zeroth sensor. Use a slice to select a range of values, e.g.,
        ``obj.sel(frequency=slice(10, 100))``.

        For more details, see `xarray.DataArray.isel` and `xarray.Dataset.isel`.
        """
        new = self.data.isel(indexers=indexers, drop=drop, missing_dims=missing_dims, **indexers_kwargs)
        if drop_allnan:
            new = new.where(~new.isnull(), drop=True)
        return self.__array_wrap__(new)

    @property
    def coords(self):
        """The coordinate (dimension) arrays for this data.

        Refer to `xarray.DataArray.coords` and `xarray.Dataset.coords`.
        """
        return self.data.coords

    @property
    def dims(self):
        """The dimensions of this data.

        Refer to `xarray.DataArray.dims` and `xarray.Dataset.dims`.
        """
        return self.data.dims

    def groupby(self, group):
        for label, group in self.data.groupby(group, squeeze=False):
            yield label, self.__array_wrap__(group.squeeze())

    def _figure_template(self, **kwargs):
        """Create default figure layout for this data."""
        import plotly.graph_objects as go
        import plotly.io as pio

        template = go.layout.Template()
        template.update(pio.templates[pio.templates.default])
        return template

    def make_figure(self, **kwargs):
        """Create a plotly figure, styled for this data.

        Some useful keyword arguments:

        - ``xaxis_title`` and ``yaxis_title`` controls axis titles for the figure.
        - ``height`` and ``width`` sets the figure size in pixels.
        - ``title`` adds a top level title.
        """
        import plotly.graph_objects as go

        fig = go.Figure(layout=dict(template=self._figure_template(**kwargs), **kwargs))
        return fig

    def save(self, path, append_dim=None, **kwargs):
        """Save the data to a Zarr file at the specified path.

        The class name is stored as an attribute, so it can be used later to reconstruct the object.

        Parameters
        ----------
        path : str or pathlib.Path
            The file path where the data will be saved. The directory will be
            created if it doesn't exist.
        append_dim : str, optional
            A dimension which should be used to append to existing data in the path.
        **kwargs : dict, optional
            Additional keyword arguments passed to the `xarray.Dataset.to_zarr` method, which
            is responsible for saving the data to the Zarr format.
        """
        from pathlib import Path

        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        module = self.__class__.__module__
        name = self.__class__.__qualname__
        if module is not None and module != "__builtin__":
            name = module + "." + name

        data = self.data.assign_attrs(__uwacan_class__=name)

        if append_dim:
            if append_dim not in data.dims:
                data = data.expand_dims(append_dim)
            if not path.exists():
                # Cannot append if there is no data to append to, so we just write as normal
                append_dim = None
        if isinstance(data, xr.Dataset):
            for var in data.dara_vars.values():
                if np.issubdtype(var.dtype, np.datetime64):
                    var.encoding = {"units": "nanoseconds since 1970-01-01"}
        for coord in data.coords.values():
            if np.issubdtype(coord.dtype, np.datetime64):
                coord.encoding = {"units": "nanoseconds since 1970-01-01"}

        kwargs.setdefault("consolidated", False)
        data.to_zarr(path, append_dim=append_dim, **kwargs)

    @classmethod
    def load(cls, path, lookup_class=True, **kwargs):
        """Load data from a Zarr file and optionally restore the original class.

        This method loads data from a Zarr file and attempts to reconstruct the
        original class that was used to save the data. The class information is
        stored in the `__uwacan_class__` attribute of the dataset. If the class
        is found, the method dynamically loads and instantiates it.

        Parameters
        ----------
        path : str or pathlib.Path
            The file path from which to load the data.
        lookup_class : bool, default=True
            If True attempts to restore the original class from the
            metadata stored in the Zarr file. If False, the called class is used
            to load the data.

        Returns
        -------
        obj : cls or ``__uwacan_class__``
            An instance of the class used to save the data (if found in the Zarr
            file's metadata), or an instance of the called class.

        Notes
        -----
        - The Zarr file should have the `__uwacan_class__` attribute in the
          dataset's metadata to allow class reconstruction.
        - If the class cannot be found, or if there is an error during dynamic
          import, the method falls back to using the current class `cls` to
          instantiate the object.

        """
        kwargs.setdefault("consolidated", False)
        data = xr.open_zarr(path, **kwargs)
        if "__xarray_dataarray_variable__" in data:
            data = data["__xarray_dataarray_variable__"]
            data.name = None
        if lookup_class and "__uwacan_class__" in data.attrs:
            import importlib

            module_name, class_name = data.attrs["__uwacan_class__"].rsplit(".", 1)
            try:
                module = importlib.import_module(module_name)
                cls = getattr(module, class_name)
            except:
                pass
        return cls(data)


class DataArrayWrap(xrwrap, np.lib.mixins.NDArrayOperatorsMixin):
    """Wrapper around `xarray.DataArray`.

    This base class exists to wrap functionality in `xarray.DataArray`,
    and numerical operations from `numpy`.
    """

    _coords_set_by_init = set()

    def __init__(self, data, dims=(), coords=None, **kwargs):
        if isinstance(data, DataArrayWrap):
            data = data.data
        if not isinstance(data, xr.DataArray):
            if isinstance(dims, str):
                dims = [dims]
            if dims is None:
                dims = ()
            if np.ndim(data) != np.size(dims):
                raise ValueError(
                    f"Dimension names '{dims}' for {type(self).__name__} does not match data with {np.ndim(data)} dimensions"
                )
            data = xr.DataArray(data, dims=dims)
        if coords is not None:
            data = data.assign_coords(
                **{name: coord for (name, coord) in coords.items() if name not in self._coords_set_by_init}
            )
        super().__init__(data, **kwargs)

    def __array__(self, dtype=None):
        """Casts this object into a `numpy.ndarray`."""
        return self.data.__array__(dtype)

    @staticmethod
    def _implements_np_func(np_func):
        """Tag implementations of `numpy` functions.

        We use the ``__array_function__`` interface to implement many
        `numpy` functions. This decorator will only tag an implementation
        function with which `numpy` function it implements.
        """

        def decorator(func):
            func._implements_np_func = np_func
            return func

        return decorator

    def __init_subclass__(cls):
        """Set up the `numpy` implementations for a class.

        This will run when a subclass is defined, and
        check if there are any methods in it that are tagged
        with a numpy implementation. All those implementations
        will be stored in a class-level dictionary.
        """
        implementations = {}
        for name, value in cls.__dict__.items():
            if callable(value) and hasattr(value, "_implements_np_func"):
                implementations[value._implements_np_func] = value
        cls._np_func_implementations = implementations

    def __array_function__(self, func, types, args, kwargs):
        """Interfaces with numpy functions.

        This will run when general numpy functions are used on objects
        of this class. We have stored tagged implementations in class
        dictionaries, so we can check if there is an explicit implementation.
        We have no actual method which does this, so we go through the ``mro``
        manually.

        If no explicit implementation is found, we try replacing all wrappers
        with their `xarray.DataArray` objects, and call the function on them
        instead.
        """
        for cls in self.__class__.mro():
            if hasattr(cls, "_np_func_implementations"):
                if func in cls._np_func_implementations:
                    func = cls._np_func_implementations[func]
                    break
        else:
            # We couldn't find an explicit implementation.
            # Try replacing all _DataWrapper with their data and calling the function.
            args = (arg.data if isinstance(arg, DataArrayWrap) else arg for arg in args)
            out = func(*args, **kwargs)
            if not isinstance(out, xr.DataArray):
                try:
                    out = self.data.__array_wrap__(out)
                except:
                    # We cannot wrap this in an xarray, then we cannot wrap in our own wrapper.
                    return out
            out = self.__array_wrap__(out)
            return out
        return func(*args, **kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Interfaces with `numpy.ufunc`.

        Many functions in numpy are ufuncs. `xarray.DataArray.__array_ufunc__` will
        do the heavy lifting here.
        """
        inputs = (arg.data if isinstance(arg, DataArrayWrap) else arg for arg in inputs)
        return self.__array_wrap__(self.data.__array_ufunc__(ufunc, method, *inputs, **kwargs))

    @_implements_np_func(np.mean)
    def mean(self, dim=..., **kwargs):
        """Average of this data, along some dimension.

        See `xarray.DataArray.mean` for more details.
        """
        return self.__array_wrap__(self.data.mean(dim, **kwargs))

    @_implements_np_func(np.sum)
    def sum(self, dim=..., **kwargs):
        """Sum of this data, along some dimension.

        See `xarray.DataArray.sum` for more details.
        """
        return self.__array_wrap__(self.data.sum(dim, **kwargs))

    @_implements_np_func(np.std)
    def std(self, dim=..., **kwargs):
        """Standard deviation of this data, along some dimension.

        See `xarray.DataArray.std` for more details.
        """  # noqa: D401
        return self.__array_wrap__(self.data.std(dim, **kwargs))

    @_implements_np_func(np.max)
    def max(self, dim=..., **kwargs):
        """Maximum of this data, along some dimension.

        See `xarray.DataArray.max` for more details.
        """
        return self.__array_wrap__(self.data.max(dim, **kwargs))

    @_implements_np_func(np.min)
    def min(self, dim=..., **kwargs):
        """Minimum of this data, along some dimension.

        See `xarray.DataArray.min` for more details.
        """
        return self.__array_wrap__(self.data.min(dim, **kwargs))

    def apply(self, func, *args, **kwargs):
        """Apply some function to the contained data.

        This calls the supplied function with the `xarray.DataArray`
        in this object, then wraps the output in a similar container again.
        """
        data = func(self.data, *args, **kwargs)
        return self.__array_wrap__(data)

    def reduce(self, func, dim, **kwargs):
        """Apply a reduction function along some dimension in this data.

        See `xarray.DataArray.reduce` for more details.
        """
        data = self.data.reduce(func=func, dim=dim, **kwargs)
        return self.__array_wrap__(data)


DataArrayWrap.__init_subclass__()


class DatasetWrap(xrwrap):
    """Wraps `xarray.Dataset` objects.

    This wraps a dataset by passing indexing to the underlying dataset
    indexing, and mimics the `xarray` attribute access by passing
    attribute access to indexing if the attribute exists in the dataset.
    Using a MutableMapping from collections enables lots of dict-style
    iteration.
    """

    def __getitem__(self, key):
        try:
            return self._data[key]
        except KeyError as e:
            raise KeyError(*e.args) from None

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        return key in self.data

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def values(self):
        return self.data.values()

    def get(self, key, default=None):
        return self.data.get(key, default=default)

    def __getattr__(self, key):
        try:
            return self._data[key]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'") from None

    def __dir__(self):
        return sorted(set(super().__dir__()) | set(self._data.variables))


class Roller:
    """Base class for rolling windows."""

    @property
    def num_frames(self):
        """The number of frames in this rolling output."""
        return self.settings["num_frames"]

    @property
    def shape(self):
        """The shape of the output frames."""
        ...

    @property
    def dims(self):
        """The dimensions of the output frames."""
        ...

    @property
    def coords(self):
        """The coords of the output frames."""
        ...

    def numpy_frames(self):
        """Generate numpy frames with data."""
        yield

    def __iter__(self):
        """Generate rolling frames of the contained data, of the same type."""
        yield


class TimeData(DataArrayWrap):
    """Handing data which varies over time.

    This class is mainly used to wrap time-signals of sampled sounds.
    As such, the time data is assumed to be sampled at a constant samplerate.

    Parameters
    ----------
    data : array_like
        A `numpy.ndarray` or a `xarray.DataArray` with the time data.
    time : array_like, optional
        A `numpy.ndarray` with ``dtype=datetime64[ns]`` containing time stamps for the samples.
    start_time : time_like, optional
        The start time for the first sample in the signal.
        This should ideally be a proper time type, but it will be parsed if it is a string.
        Defaults to "now" if not given.
    samplerate : float, optional
        The samplerate for this data, in Hz.
        If the ``data`` is a `numpy.ndarray`, this has to be given.
        If the ``data`` is a `xarray.DataArray` which already has a time coordinate,
        this can be omitted.
    dims : str or [str], default="time"
        The dimensions of the data. Must have the same length as the number of dimensions in the data.
        Only used for `numpy` inputs.
    coords : `xarray.DataArray.coords`
        Additional coordinates for this data.
    attrs : dict, optional
        Additional attributes to store with this data.
    """

    _coords_set_by_init = {"time"}

    def __init__(
        self, data, time=None, start_time=None, samplerate=None, dims="time", coords=None, attrs=None, **kwargs
    ):
        super().__init__(data, dims=dims, coords=coords, attrs=attrs, **kwargs)

        if samplerate is None and time is not None:
            samplerate = np.timedelta64(1, "s") / np.mean(np.diff(time[:1000]))
        elif time is None and samplerate is not None:
            if start_time is None:
                if "time" in self.data.coords:
                    start_time = self.data.time[0].item()
                else:
                    start_time = "now"
            n_samples = self.data.sizes["time"]
            start_time = time_to_np(start_time)
            offsets = np.arange(n_samples) * 1e9 / samplerate
            time = start_time + offsets.astype("timedelta64[ns]")

        if time is not None:
            if not isinstance(time, xr.DataArray):
                time = xr.DataArray(time, dims="time")
            if samplerate is not None:
                time.attrs["rate"] = samplerate
            self.data.coords["time"] = time
        elif "rate" not in self.data.time.attrs:
            if samplerate is None:
                samplerate = np.timedelta64(1, "s") / np.mean(np.diff(self.data.time[:1000]))
            self.data.time.attrs["rate"] = samplerate

    @classmethod
    def from_dataset(cls, dataset):
        if "time" in dataset.dims:
            return super().from_dataset(dataset)
        # This is not time data any more, let the caller catch the error.
        raise NotImplementedError(f"Cannot instantiate '{cls.__name__}' from data lacking time dimension")

    @property
    def samplerate(self):
        return self.data.time.rate

    @property
    def time(self):
        """Time coordinates for this data."""
        return self.data.time

    @property
    def time_window(self):
        """A `TimeWindow` describing when the data start and stops."""
        # Calculating duration from number and rate means the stop points to the sample after the last,
        # which is more intuitive when considering signal durations etc.
        return TimeWindow(
            start=self.data.time.data[0],
            duration=self.data.sizes["time"] / self.samplerate,
        )

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        """Select a subset of the data over time.

        See `TimeWindow.subwindow` for details on the parameters.
        """
        original_window = self.time_window
        new_window = original_window.subwindow(
            time, start=start, stop=stop, center=center, duration=duration, extend=extend
        )
        if isinstance(new_window, TimeWindow):
            start = (new_window.start - original_window.start).in_seconds()
            stop = (new_window.stop - original_window.start).in_seconds()
            # Indices assumed to be seconds from start
            start = int(np.floor(start * self.samplerate))
            stop = int(np.ceil(stop * self.samplerate))
            idx = slice(start, stop)
        else:
            idx = (new_window - original_window.start).in_seconds()
            idx = round(idx * self.samplerate)

        selected_data = self.data.isel(time=idx)
        new = type(self)(selected_data)
        return new

    def listen(self, downsampling=1, upsampling=None, headroom=6, **kwargs):
        """Play back this time data over speakers.

        This tries to play the time data as audio using the `sounddevice` package.
        The audio will be centered at 0 and normalized before playback.

        Parameters
        ----------
        downsampling : float, optional
            Artificially uses a lower samplerate in playback to slow
            down the audio, lowering the pitch of the content.
        upsampling : int, optional
            Decimates the data by selecting every Nth sample, speeding
            up the audio and raising the pitch of the content.
        headroom : float, default 6
            How much headroom to leave in the normalization, in dB.
        **kwargs : dict, optional
            Remaining keyword arguments are passed to `sounddevice.play`.
            The most useful arguments are ``blocking=True``, and ``device``.
        """
        import sounddevice as sd

        sd.stop()
        data = self.data
        if upsampling:
            data = data[::upsampling]
        scaled = data - data.mean()
        scaled = scaled / np.max(np.abs(scaled)) * 10 ** (-headroom / 20)
        sd.play(scaled, samplerate=round(self.samplerate / downsampling), **kwargs)

    def rolling(self, duration=None, step=None, overlap=None, squeeze_time=True):
        """Generate rolling windows of this data.

        Parameters
        ----------
        duration : float
            The duration of each frame, in seconds.
        step : float
            The step between consecutive frames, in seconds.
        overlap : float
            The overlap between consecutive frames, as a fraction of the duration.
        squeeze_time : bool, default `True`
            If this is set to `False`, rolling over windows with single time values will still
            give output with a time axis/dim.

        Returns
        -------
        roller : `TimeDataRoller`
            A roller object to roll over the data.
        """
        return TimeDataRoller(self, duration=duration, step=step, overlap=overlap, squeeze_time=squeeze_time)


class TimeDataRoller(Roller):
    """Rolling windows of time data.

    Parameters
    ----------
    obj : TimeData
        The time data wrapper to roll over.
    duration : float
        The duration of each frame, in seconds.
    step : float
        The step between consecutive frames, in seconds.
    overlap : float
        The overlap between consecutive frames, as a fraction of the duration.
    squeeze_time : bool, default `True`
        If this is set to `False`, rolling over windows with single time values will still
        give output with a time axis/dim.
    """

    def __init__(self, obj, duration=None, step=None, overlap=0, squeeze_time=True):
        self.obj = obj
        self.settings = time_frame_settings(
            duration=duration,
            step=step,
            overlap=overlap,
            signal_length=self.obj.time_window.duration,
            samplerate=self.obj.samplerate,
        )

        if self.settings["samples_per_frame"] == 1 and squeeze_time:
            self._squeeze_time = True
            self._slices = list(range(self.settings["num_frames"]))
        else:
            self._squeeze_time = False
            self._slices = [
                slice(start_idx, start_idx + self.settings["samples_per_frame"])
                for start_idx in range(
                    0, self.settings["num_frames"] * self.settings["sample_step"], self.settings["sample_step"]
                )
            ]

    @property
    def coords(self):
        coords = dict(self.obj.coords)
        if not self._squeeze_time:
            coords["time"] = coords["time"][: self.settings["samples_per_frame"]]
        else:
            del coords["time"]
        return coords

    @property
    def dims(self):
        dims = list(self.obj.dims)
        dims.remove("time")
        if not self._squeeze_time:
            dims = ["time"] + dims
        return tuple(dims)

    @property
    def shape(self):
        shape = list(self.obj.data.shape)
        del shape[self.obj.dims.index("time")]
        if not self._squeeze_time:
            shape = [self.settings["samples_per_frame"]] + shape
        return tuple(shape)

    def numpy_frames(self):
        data = self.obj.data.transpose("time", ...)
        # For large datasets the data is likely disk-mapped. Load in chunks to avoid running out of memory.
        num_chunks = max(1, data.nbytes // (2**30))  # At least 1 chunk
        chunk_size = len(self._slices) // num_chunks  # Number of frames per chunk

        for i in range(0, len(self._slices), chunk_size):
            chunk_slices = self._slices[i:i + chunk_size]
            # Load the chunk of data that covers all slices in this chunk
            if isinstance(chunk_slices[0], slice):
                start_idx = chunk_slices[0].start
                end_idx = chunk_slices[-1].stop
            else:
                start_idx = chunk_slices[0]
                end_idx = chunk_slices[-1] + self.settings["samples_per_frame"]
            chunk_data = data.isel(time=slice(start_idx, end_idx)).data
            # Yield individual frames from the loaded chunk
            for s in chunk_slices:
                if isinstance(s, slice):
                    yield chunk_data[s.start - start_idx:s.stop - start_idx]
                else:
                    yield chunk_data[s - start_idx]

    def time_data(self):
        """Generate rolling frames of the contained data, as `TimeData`."""
        for s in self._slices:
            yield self.obj.isel(time=s)

    def __iter__(self):
        yield from self.time_data()


class FrequencyData(DataArrayWrap):
    """Handing data which varies over frequency.

    This class is mainly used to wrap spectra of sampled sounds.
    Typically, this is represented as power spectral densities,
    but other data types are also possible.

    Parameters
    ----------
    data : array_like
        A `numpy.ndarray` or a `xarray.DataArray` with the frequency data.
    frequency : array_like, optional
        The frequencies corresponding to the data. Mandatory if ``data`` is a `numpy.ndarray`.
    bandwidth : array_like, optional
        The bandwidth of each data point. Can be an array with per-frequency
        bandwidth or a single value valid for all frequencies.
    dims : str or [str], default="frequency"
        The dimensions of the data. Must have the same length as the number of dimensions in the data.
        Only used for `numpy` inputs.
    coords : `xarray.DataArray.coords`
        Additional coordinates for this data.
    attrs : dict, optional
        Additional attributes to store with this data.
    """

    _coords_set_by_init = {"frequency", "bandwidth"}

    def __init__(self, data, frequency=None, bandwidth=None, dims="frequency", coords=None, attrs=None, **kwargs):
        super().__init__(data, dims=dims, coords=coords, attrs=attrs, **kwargs)
        if frequency is not None:
            self.data.coords["frequency"] = frequency
        if bandwidth is not None:
            bandwidth = np.broadcast_to(bandwidth, np.shape(frequency))
            self.data.coords["bandwidth"] = ("frequency", bandwidth)

    @classmethod
    def from_dataset(cls, dataset):
        if "frequency" in dataset.dims:
            return super().from_dataset(dataset)
        # This is not frequency data any more, just return the plain xr.DataArray
        raise NotImplementedError(f"Cannot instantiate '{cls.__name__}' from data lacking frequency dimension")

    @property
    def frequency(self):
        """The frequencies for the data."""
        return self.data.frequency

    def estimate_bandwidth(self):
        """Estimate the bandwidth of the frequency vector.

        This tries to determine if the frequency vector is linearly
        or logarithmically spaced, then uses either linear or logarithmic
        bandwidth estimation.

        Returns
        -------
        bandwidth : `xarray.DataArray`
            The estimated bandwidth.
        """
        frequency = np.asarray(self.frequency)
        # Check if the frequency array seems linearly or logarithmically spaced
        if frequency[0] == 0:
            diff = frequency[2:] - frequency[1:-1]
            frac = frequency[2:] / frequency[1:-1]
        else:
            diff = frequency[1:] - frequency[:-1]
            frac = frequency[1:] / frequency[:-1]
        diff_err = np.std(diff) / np.mean(diff)
        frac_err = np.std(frac) / np.mean(frac)
        # Note: if there are three values and the first is zero, the std is 0 for both.
        # The equals option makes us end up in the linear frequency case.
        if diff_err <= frac_err:
            # Central differences, with forwards and backwards at the ends
            central = (frequency[2:] - frequency[:-2]) / 2
            first = frequency[1] - frequency[0]
            last = frequency[-1] - frequency[-2]
        else:
            # upper edge is at sqrt(f_{l+1} * f_l), lower edge is at sqrt(f_{l-1} * f_l)
            # the difference simplifies as below.
            central = (frequency[2:] ** 0.5 - frequency[:-2] ** 0.5) * frequency[1:-1] ** 0.5
            # extrapolating to one bin below lowest and one above highest using constant ratio
            # the expression above then simplifies to the expressions below
            first = (frequency[1] - frequency[0]) * (frequency[0] / frequency[1]) ** 0.5
            last = (frequency[-1] - frequency[-2]) * (frequency[-1] / frequency[-2]) ** 0.5
        bandwidth = np.concatenate([[first], central, [last]])
        return xr.DataArray(bandwidth, coords={"frequency": self.frequency})

    def _figure_template(self, **kwargs):
        template = super()._figure_template(**kwargs)
        template.layout.update(
            xaxis=dict(
                type="log",
                title="Frequency in Hz",
            )
        )
        return template

    def plot(self, **kwargs):
        """Make a scatter trace of this data.

        Some useful keyword arguments:

        - ``name`` sets the legendentry of this trace.
        - ``line_color`` and ``marker_color`` sets the line and marker colors.
        - ``mode`` sets the mode, typically one of ``"lines"``, ``"markers"``, or ``"markers+lines"``.

        """
        import plotly.graph_objects as go

        if self.dims != ("frequency",):
            raise ValueError(
                f"Cannot simply scatter frequency data with dimensions '{self.dims}'. "
                "Use the `.groupby(dim)` method to loop over non-frequency dimensions."
            )

        return go.Scatter(
            x=self.frequency,
            y=self.data,
            **kwargs,
        )


class TimeFrequencyData(TimeData, FrequencyData):
    """Handing data which varies over time and frequency.

    This class is mainly used to wrap spectrogram-like data.
    There are no processing implemented in this class, but
    subclasses are free to add processing methods, custom
    initialization, or instantiation in class methods.

    Parameters
    ----------
    data : array_like
        A `numpy.ndarray` or a `xarray.DataArray` with the time-frequency data.
    start_time : time_like, optional
        The start time for the first sample in the signal.
        This should ideally be a proper time type, but it will be parsed if it is a string.
        Defaults to "now" if not given.
    samplerate : float, optional
        The samplerate for this data, in Hz. This is not the samplerate of the underlying time signal,
        but time steps of the time axis on the time-frequency data.
        If the `data` is a `numpy.ndarray`, this has to be given.
        If the `data` is a `xarray.DataArray` which already has a time coordinate,
        this can be omitted.
    frequency : array_like, optional
        The frequencies corresponding to the data. Mandatory if `data` is a `numpy.ndarray`.
    bandwidth : array_like, optional
        The bandwidth of each data point. Can be an array with per-frequency
        bandwidth or a single value valid for all frequencies.
    dims : str or [str], optional
        The dimensions of the data. Must have the same length as the number of dimensions in the data.
        Mandatory used for `numpy` inputs, not used for `xarray` inputs.
    coords : `xarray.DataArray.coords`
        Additional coordinates for this data.
    attrs : dict, optional
        Additional attributes to store with this data.
    """

    _coords_set_by_init = {"time", "frequency", "bandwidth"}

    def __init__(
        self,
        data,
        start_time=None,
        samplerate=None,
        frequency=None,
        bandwidth=None,
        dims=None,
        coords=None,
        attrs=None,
        **kwargs,
    ):
        super().__init__(
            data,
            dims=dims,
            coords=coords,
            attrs=attrs,
            start_time=start_time,
            samplerate=samplerate,
            frequency=frequency,
            bandwidth=bandwidth,
            **kwargs,
        )

    @classmethod
    def from_dataset(cls, data, **kwargs):
        if "frequency" not in data.dims:
            # It's not frequency-data, but it might be time data
            return TimeData.from_dataset(data, **kwargs)
        if "time" not in data.dims:
            # It's not time-data, but it might be frequency data
            return FrequencyData.from_dataset(data, **kwargs)
        return super().from_dataset(data, **kwargs)

    def rolling(self, duration=None, step=None, overlap=None, squeeze_time=True):
        """Generate rolling windows of this data.

        By default, a single time instance of the data will be
        generated in each frame.

        Parameters
        ----------
        duration : float
            The duration of each frame, in seconds.
        step : float
            The step between consecutive frames, in seconds.
        overlap : float
            The overlap between consecutive frames, as a fraction of the duration.
        squeeze_time : bool, default `True`
            If this is set to `False`, rolling over windows with single time values will still
            give output with a time axis/dim.

        Returns
        -------
        roller : `TimeDataRoller`
            A roller object to roll over the data.
        """
        if (duration, step) == (None, None):
            step = 1 / self.samplerate
            overlap = 0
        return TimeDataRoller(self, duration=duration, step=step, overlap=overlap, squeeze_time=squeeze_time)

    def _figure_template(self, **kwargs):
        template = super()._figure_template(**kwargs)
        template.layout.update(
            xaxis=dict(
                title_text="Time",
                type=None,  # Use default which is `linear`` for "normal" data and `date` for timestamp data
            ),
            yaxis=dict(
                title_text="Frequency in Hz",
                type="log",
            ),
        )
        template.data.update(
            heatmap=[
                dict(
                    colorscale="viridis",
                    colorbar_title_side="right",
                    hovertemplate="%{x}<br>%{y:.5s}Hz<br>%{z}<extra></extra>",
                )
            ]
        )
        return template

    def plot(self, **kwargs):
        """Make a heatmap trace of this data.

        Some useful keyword arguments:

        - ``name`` sets the legendentry of this trace.
        - ``colorscale`` chooses the colorscale, e.g., ``"viridis"``, ``"delta"``, ``"twilight"``.
        - ``zmin`` and ``zmax`` sets the color range.

        """
        import plotly.graph_objects as go

        if set(self.dims) != {"frequency", "time"}:
            raise ValueError(
                f"Cannot make heatmap of  time-frequency data with dimensions '{self.dims}'. "
                "Use the `.groupby(dim)` method to loop over extra dimensions."
            )

        trace = go.Heatmap(
            x=self.time,
            y=self.frequency,
            z=self.data.transpose("frequency", "time"),
        )
        return trace.update(**kwargs)


class Transit:
    """Container for recorded ship transits.

    This class is responsible for bundling recordings and position tracks.
    Note that this class does not take `TimeData` as the input.
    The track and recording will be restricted to a time window which
    both of them covers.

    Attributes
    ----------
    recording : `recordings.Recording`
        A recording of the ship transit.
    track : `positional.Track`
        The position track of the ship.
    """

    def __init__(self, recording, track):
        start = max(recording.time_window.start, track.time_window.start)
        stop = min(recording.time_window.stop, track.time_window.stop)

        self.recording = recording.subwindow(start=start, stop=stop)
        self.track = track.subwindow(start=start, stop=stop)

    @property
    def time_window(self):
        """A `TimeWindow` describing when the data start and stops."""
        rec_window = self.recording.time_window
        track_window = self.track.time_window
        return TimeWindow(start=max(rec_window.start, track_window.start), stop=min(rec_window.stop, track_window.stop))

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        """Select a subset of the data over time.

        See `TimeWindow.subwindow` for details on the parameters.
        """
        subwindow = self.time_window.subwindow(
            time, start=start, stop=stop, center=center, duration=duration, extend=extend
        )
        rec = self.recording.subwindow(subwindow)
        track = self.track.subwindow(subwindow)
        return type(self)(recording=rec, track=track)


def dB(x, power=True, safe_zeros=True, ref=1):
    """Calculate the decibel of an input value.

    Parameters
    ----------
    x : numeric
        The value to take the decibel of
    power : boolean, default True
        Specifies if the input is a power-scale quantity or a root-power quantity.
        For power-scale quantities, the output is 10 log(x), for root-power quantities the output is 20 log(x).
        If there are negative values in a power-scale input, the handling can be controlled as follows:
        - `power='imag'`: return imaginary values
        - `power='nan'`: return nan where power < 0
        - `power=True`: as `nan`, but raises a warning.
    safe_zeros : boolean, default True
        If this option is on, all zero values in the input will be replaced with the smallest non-zero value.
    ref : numeric
        The reference unit for the decibel. Note that this should be in the same unit as the `x` input,
        e.g., if `x` is a power, the `ref` value might need squaring.
    """
    if isinstance(x, DataArrayWrap):
        return x.apply(dB, power=power, safe_zeros=safe_zeros, ref=ref)

    if safe_zeros and np.size(x) > 1:
        nonzero = x != 0
        min_value = np.nanmin(abs(xr.where(nonzero, x, np.nan, keep_attrs=True)))
        x = xr.where(nonzero, x, min_value, keep_attrs=True)
    if power:
        if np.any(x < 0):
            if power == "imag":
                return 10 * np.log10(x + 0j)
            if power == "nan":
                return 10 * np.log10(xr.where(x > 0, x, np.nan, keep_attrs=True))
        return 10 * np.log10(x / ref)
    else:
        return 20 * np.log10(np.abs(x) / ref)


def time_frame_settings(
    duration=None,
    step=None,
    overlap=None,
    resolution=None,
    num_frames=None,
    signal_length=None,
    samplerate=None,
):
    """Calculate time frame overlap settings from various input parameters.

    Parameters
    ----------
    duration : float
        How long each frame is, in seconds
    step : float
        The time between frame starts, in seconds
    overlap : float
        How much overlap there is between the frames, as a fraction of the duration.
        If this is negative the frames will have extra space between them.
    resolution : float
        Desired frequency resolution in Hz. Equals ``1/duration``
    num_frames : int
        The total number of frames in the signal
    signal_length : float, optional
        The total length of the signal, in seconds
    samplerate : float, optional
        The samplerate of the signal, in Hz.
        Only used to compute sample versions of the outputs.

    Returns
    -------
    settings : dict
        The settings dict has keys:
            ``"duration"``, ``"step"``, ``"overlap"``, ``"resolution"``
        and if ``signal_length`` was given:
            ``"num_frames"``, ``"signal_length"``
        and if ``samplerate`` was given:
            ``"samples_per_frame"``, ``"sample_step"``, ``"sample_overlap"``
        and if both ``samplerate`` and ``signal_length`` was given:
            ``"sample_total"``

    Raises
    ------
    ValueError
        If the inputs are not sufficient to determine the frame,
        or if priorities cannot be disambiguated.

    Notes
    -----
    The parameters will be used in the following priority:

    1. ``signal_length`` and ``num_frames``
    2. ``step`` and ``duration``
    3. ``resolution`` and ``overlap``

    Each frame ``idx=[0, ..., num_frames - 1]`` has::

        start = idx * step
        stop = idx * step + duration

    The last frame thus ends at ``(num_frames - 1) step + duration``.
    The overlap relations are::

        duration = step / (1 - overlap)
        step = duration (1 - overlap)
        overlap = 1 - step / duration

    If both ``resolution`` and ``overlap`` are given, they will be treated as "minimum"
    parameters. This means that the output will have at least an overlap as specified
    (`overlap_out >= overlap_in`) and at least the specified spectral resolution
    (`resolution_out <= resolution_in`). If the other input parameters are sufficient
    to fully determine the frame settings, the output will be checked to meet the
    requested resolution and overlap criteria.
    The overlap will default to a minimum of 0 if not given, i.e. frames that are
    edge to edge (unless more overlap is required from other parameters).

    This gives us the following total list of priorities:

    1) ``signal_length`` and ``num_frames`` are given
        a) ``step`` or ``duration`` (not both!) given
        b) ``resolution`` is given
        c) ``overlap`` is given
    2) ``step`` and ``duration`` given
    3) ``step`` given
        a) ``resolution`` is given
        b) ``overlap`` is given
    4) ``duration`` given (``resolution`` is ignored, ``overlap`` is used)
    5) ``resolution`` given

    For cases 2-5, ``num_frames`` is calculated if ``signal_length`` is given,
    and a new truncated ``signal_length`` is returned.

    If a samplerate was given, the number of samples in an output is also computed.
    This is done with::

        samples_per_frame = ceil(duration * samplerate)
        sample_step = floor(step * samplerate)
        sample_overlap = samples_per_frame - sample_step

    We use ceil for the duration since it is often chosen to obtain a minimum frequency resolution.
    We use floor for the step to make sure we fit all the frames in the total length.
    Note that this means that the sample overlap might not equal
    ``round(duration * overlap * samplerate)``, due to how the rounding is done.
    It can at most be two samples larger than the rounded value.
    """
    if None not in (num_frames, signal_length):
        if None not in (duration, step):
            raise ValueError(
                "Overdetermined time frames: "
                "All four of `num_frames`, `signal_length`, `duration`, and `step` "
                "cannot be specified simultaneously."
            )
        elif step is not None:
            # We have the step, calculate the duration
            duration = signal_length - (num_frames - 1) * step
        elif duration is not None:
            # We have the duration, calculate the step
            step = (signal_length - duration) / (num_frames - 1)
        elif resolution is not None:
            duration = 1 / resolution
            step = (signal_length - duration) / (num_frames - 1)
            overlap = 1 - step / duration
        else:
            # This sets a duration to meet a minimum overlap, at least 0.
            duration = signal_length / (num_frames + (overlap or 0) - num_frames * (overlap or 0))
            if resolution is not None:
                # If a resolution requires longer duration, that's set here.
                duration = max(duration, 1 / resolution)
            step = (signal_length - duration) / (num_frames - 1)
    elif None in (step, duration):
        # Missing one of step and duration, infer it from overlap and resolution
        if duration is not None:
            step = duration * (1 - (overlap or 0))
        elif step is not None:
            duration = step / (1 - (overlap or 0))
            if resolution is not None:
                duration = max(duration, 1 / resolution)
        elif resolution is not None:
            duration = 1 / resolution
            step = duration * (1 - (overlap or 0))
        else:
            raise ValueError(
                "Must give at least one of (`step`, `duration`, `resolution`) or the pair of `signal_length` and `num_frames`."
            )

    if overlap is not None:
        if (1 - step / duration < overlap) and not np.isclose(1 - step / duration, overlap):
            raise ValueError(f"Time frame step {step}s and duration {duration}s does not meet minimum overlap fraction {overlap}")
    if resolution is not None:
        if (1 / duration > resolution) and not np.isclose(1 / duration, resolution):
            raise ValueError(f"Duration {duration}s does not meet minimum spectral resolution {resolution}Hz.")

    settings = {
        "duration": duration,
        "step": step,
        "overlap": 1 - step / duration,
        "resolution": 1 / duration,
    }
    if signal_length is not None:
        num_frames = num_frames or int(np.floor((signal_length - duration) / step + 1))
        settings["num_frames"] = num_frames
        settings["signal_length"] = (num_frames - 1) * step + duration
    if samplerate is not None:
        settings["samples_per_frame"] = int(np.ceil(samplerate * duration))
        settings["sample_step"] = int(np.floor(samplerate * step))
        settings["sample_overlap"] = settings["samples_per_frame"] - settings["sample_step"]
        if "num_frames" in settings:
            settings["sample_total"] = (settings["num_frames"] - 1) * settings["sample_step"] + settings["samples_per_frame"]  # fmt: skip
    return settings


def concatenate(items, dim=None, nan_between_items=False, sort=False, cls=None, **kwargs):
    """Concatenates a list of items along a specified dimension.

    Parameters
    ----------
    items : list of `xarray.Dataset` or `xarray.DataArray` or `xrwrap`
        A list of items to be concatenated. The items can be `xr.Dataset`,
        `xr.DataArray`, or instances of a subclass of `xrwrap`.
        Most `uwacan` data wrappers are compatible.
    dim : str or None, optional
        The dimension along which to concatenate. If `None`, the function will
        attempt to infer the dimension from the first item. If the items have
        more than one dimension, a `ValueError` will be raised.
        The concatenation dimension can be an existing dim or a new dim.
    nan_between_items : bool, optional, default=False
        If `True`, inserts NaN values between concatenated items along the specified dimension.
        This is useful for visualization purposes, as it makes most plotting libraries split the lines.
    sort : bool, optional, default=False
        If `True`, sorts the concatenated items by the specified dimension.
    cls : class or callable, optional
        The class or callable to use for wrapping the concatenated result. If `None`,
        the function will try to infer the class from the first item. If `cls` is a subclass
        of `xrwrap` and has a `from_dataset` method, it will use that method to wrap
        the result.
    **kwargs : dict
        Additional keyword arguments passed to the class or callable specified by `cls` when
        wrapping the concatenated result.

    Returns
    -------
    object
        The concatenated object, either as an `xr.Dataset`, `xr.DataArray`, or as an
        instance of the specified `cls` (or inferred class).
    """
    if dim is None:
        if len(items[0].dims) != 1:
            raise ValueError("Cannot guess concatenation dimensions for multi-dimensional items.")
        dim = next(iter(items[0].dims))

    if cls is None:
        for item in items:
            if isinstance(item, xrwrap):
                cls = type(item)
                break

    if isinstance(cls, type) and issubclass(cls, xrwrap) and hasattr(cls, "from_dataset"):
        cls = cls.from_dataset

    items = [item.data if isinstance(item, xrwrap) else item for item in items]
    if nan_between_items:
        items = [x for item in items for x in [item, xr.full_like(item.isel({dim: -1}), np.nan).expand_dims(dim)]][:-1]
    items = xr.concat(items, dim=dim)
    if sort:
        items = items.sortby(dim)
    if cls:
        items = cls(items, **kwargs)
    return items



import numpy as np

WGS84_equatorial_radius = 6_378_137.0
WGS84_polar_radius = 6_356_752.3
mercator_scale_factor = 0.9996

def nm_to_m(nm):
    """Convert nautical miles to meters"""
    return nm * 1852


def m_to_nm(m):
    """Convert meters to nautical miles"""
    return m / 1852


def mps_to_knots(mps):
    """Convert meters per second to knots"""
    return mps * (3600 / 1852)


def knots_to_mps(knots):
    """Convert knots to meters per second"""
    return knots * (1852 / 3600)


def knots_to_kmph(knots):
    """Convert knots to kilometers per hour"""
    return knots * 1.852


def kmph_to_knots(kmph):
    """Convert kilometers per hour to knots"""
    return kmph / 1.852


def local_mercator_to_wgs84(easting, northing, reference_latitude, reference_longitude):
    r"""Convert local mercator coordinates into wgs84 coordinates.

    Conventions here are :math:`λ` as the longitude and :math:`φ` as the latitude,
    :math:`x` is easting and :math:`y` is northing.

    .. math::

        λ &= λ_0 + x/R \\
        φ &= 2\arctan(\exp((y + y_0)/R)) - π/2

    The northing offset :math:`y_0` is computed by converting the reference point into
    a mercator projection with the `wgs84_to_local_mercator` function, using (0, 0) as
    the reference coordinates.
    """
    radius = WGS84_equatorial_radius * mercator_scale_factor

    ref_east, ref_north = wgs84_to_local_mercator(reference_latitude, reference_longitude, 0, 0)
    longitude = reference_longitude + np.degrees(easting / radius)
    latitude = 2 * np.arctan(np.exp((northing + ref_north) / radius)) - np.pi / 2
    latitude = np.degrees(latitude)
    return latitude, longitude


def wgs84_to_local_mercator(latitude, longitude, reference_latitude, reference_longitude):
    r"""Convert wgs84 coordinates into a local mercator projection.

    Conventions here are :math:`λ` as the longitude and :math:`φ` as the latitude,
    :math:`x` is easting and :math:`y` is northing.

    .. math::
        x &= R(λ - λ_0)\\
        y &= R\ln(\tan(π/4 + (φ - φ_0)/2))
    """
    radius = WGS84_equatorial_radius * mercator_scale_factor
    local_longitude = np.radians(longitude - reference_longitude)
    local_latitude = np.radians(latitude - reference_latitude)
    easting = radius * local_longitude
    northing = radius * np.log(np.tan(np.pi / 4 + local_latitude / 2))
    return easting, northing

def wrap_angle(degrees):
    '''Wrap an angle to (-180, 180].'''
    return 180 - np.mod(180 - degrees, 360)


_re_rp_2 = (WGS84_equatorial_radius / WGS84_polar_radius)**2
def _geodetic_to_geocentric(lat):
    r"""Compute the geocentric latitude from geodetic, in radians.

    The geocentric latitude is the latitude as seen from the center
    of the earth. The geodetic latitude is the angle formed with the
    equatorial plane when drawing the normal at a surface on the earth.

    The conversion for the geocentric latitude :math:`\hat\varphi` is

    .. math::
        \hat φ = \arctan(\tan(φ) b^2/a^2)

    with the geodetic latitude :math:`φ`, and the equatorial and polar
    earth radii :math:`a, b` respectively.
    """
    return np.arctan(np.tan(lat) / _re_rp_2)


def _geocentric_to_geodetic(lat):
    r"""Compute the geodetic latitude from geocentric, in radians.

    The geodetic latitude is the angle formed with the
    equatorial plane when drawing the normal at a surface on the earth.
    The geocentric latitude is the latitude as seen from the center
    of the earth.

    The conversion for the geocentric latitude :math:`φ` is

    .. math::
         φ = \arctan(\tan(\hat φ) a^2/b^2)

    with the geocentric latitude :math:`\hat φ`, and the equatorial and polar
    earth radii :math:`a, b` respectively.
    """
    return np.arctan(np.tan(lat) * _re_rp_2)


def _local_earth_radius(lat):
    r"""Computes the earth radius at a given latitude, in radians.

    The formula is

    .. math::
        R( φ) = \sqrt{\frac{(a^2\cos φ)^2+(b^2\sin φ)^2}{(a\cos φ)^2+(b\sin φ)^2}}

    with the geodetic latitude :math:`φ`, and the equatorial and polar earth radii
    :math:`a, b` respectively, see https://en.wikipedia.org/wiki/Earth_radius#Location-dependent_radii.
    """
    return (
        ((WGS84_equatorial_radius**2 * np.cos(lat))**2 + (WGS84_polar_radius**2 * np.sin(lat))**2)
        /
        ((WGS84_equatorial_radius * np.cos(lat))**2 + (WGS84_polar_radius * np.sin(lat))**2)
    )**0.5


def _haversine(theta):
    r"""Computes the haversine of an angle, in radians.

    This is the same as

    .. math:: \sin^2(θ/2)
    """
    return np.sin(theta / 2)**2


def distance_to(lat_1, lon_1, lat_2, lon_2):
    r"""Calculate the distance between two coordinates.

    Conventions here are λ as the longitude and φ as the latitude.
    This implementation uses the Haversine formula:

    .. math::
        c &= H(Δφ) + (1 - H(Δφ) - H(φ_1 + φ_2)) ⋅ H(Δλ)\\
        d &= 2 R(φ) ⋅ \arcsin(\sqrt{c})\\
        H(θ) &= \sin^2(θ/2)

    implemented internally with conversions to geocentric coordinates
    and a latitude-dependent earth radius.

    Parameters
    ----------
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        Distance between the two points in meters.

    """
    lat_1 = np.radians(lat_1)
    lon_1 = np.radians(lon_1)
    lat_2 = np.radians(lat_2)
    lon_2 = np.radians(lon_2)
    r = _local_earth_radius((lat_1 + lat_2) / 2)
    lat_1 = _geodetic_to_geocentric(lat_1)
    lat_2 = _geodetic_to_geocentric(lat_2)
    central_angle = _haversine(lat_2 - lat_1) + (1 - _haversine(lat_1 - lat_2) - _haversine(lat_1 + lat_2)) * _haversine(lon_2 - lon_1)
    d = 2 * r * np.arcsin(central_angle ** 0.5)
    return d


def bearing_to(lat_1, lon_1, lat_2, lon_2):
    r"""Calculate the heading from one coordinate to another.

    Conventions here are λ as the longitude and φ as the latitude.
    The implementation is based on spherical trigonometry, with
    conversions to geocentric coordinates.
    This can be written as

    .. math::
        Δx &= \cos(φ_1) \sin(φ_2) - \sin(φ_1)\cos(φ_2)\cos(φ_2 - φ_1) \\
        Δy &= \sin(λ_2 - λ_1)\cos(φ_2) \\
        θ &= \arctan(Δy/Δx)

    Parameters
    ----------
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        Bearing from the first point to the second point in degrees, wrapped to (-180, 180].
    """
    lat_1 = np.radians(lat_1)
    lon_1 = np.radians(lon_1)
    lat_2 = np.radians(lat_2)
    lon_2 = np.radians(lon_2)
    lat_1 = _geodetic_to_geocentric(lat_1)
    lat_2 = _geodetic_to_geocentric(lat_2)

    dy = np.sin(lon_2 - lon_1) * np.cos(lat_2)
    dx = np.cos(lat_1) * np.sin(lat_2) - np.sin(lat_1) * np.cos(lat_2) * np.cos(lat_2 - lat_1)

    bearing = np.arctan2(dy, dx)
    return wrap_angle(np.degrees(bearing))


def shift_position(lat, lon, distance, bearing):
    r"""Shifts a position given by latitude and longitude by a certain distance and bearing.

    The implementation is based on spherical trigonometry, with internal
    conversions to geocentric coordinates, and using the local radius of the earth.
    This is expressed as

    .. math::
        φ_2 &= \arcsin(\sin(φ_1) ⋅ \cos(δ) + \cos(φ_1) ⋅ \sin(δ) ⋅ \cos(θ)) \\
        λ_2 &= λ_1 + \arctan(\frac{\sin(θ) ⋅ \sin(δ) ⋅ \cos(φ_1)}{\cos(δ) - \sin(φ_1) ⋅ \sin(φ_2)})

    where: φ is latitude, λ is longitude, θ is the bearing (clockwise from north),
    δ is the angular distance d/R; d being the distance traveled, R the earth's radius.

    Parameters
    ----------
    lat : float
        Latitude of the initial position in degrees.
    lon : float
        Longitude of the initial position in degrees.
    distance : float
        Distance to move from the initial position in meters.
    bearing : float
        Direction to move from the initial position in degrees.

    Returns
    -------
    new_lat : float
        Latitude of the new position in degrees.
    new_lon : float
        Longitude of the new position in degrees.
    """
    lat = np.radians(lat)
    lon = np.radians(lon)
    bearing = np.radians(bearing)
    r = _local_earth_radius(lat)
    lat = _geodetic_to_geocentric(lat)
    dist = distance / r  # angular distance
    new_lat = np.arcsin(np.sin(lat) * np.cos(dist) + np.cos(lat) * np.sin(dist) * np.cos(bearing))
    new_lon = lon + np.arctan2(np.sin(bearing) * np.sin(dist) * np.cos(lat), np.cos(dist) - np.sin(lat) * np.sin(new_lat))
    new_lat = _geocentric_to_geodetic(new_lat)
    return np.degrees(new_lat), np.degrees(new_lon)


def average_angle(angle, resolution=None):
    """Calculates the average angle from a list of angles and optionally rounds it to a specified resolution.

    Parameters
    ----------
    angle : array_like
        Array of angles in degrees to be averaged.
    resolution : int, str, optional
        Specifies the resolution for rounding the angle. It can be an integer specifying the number
        of divisions (e.g., 4, 8, 16) or a string ('4', '8', '16', 'four', 'eight', 'sixteen').

    Returns
    -------
    float or str
        If resolution is None, returns the average angle in degrees.
        If resolution is an integer, returns the average angle rounded to this fraction of a turn.
        If resolution is a string, returns the closest named direction (e.g., 'North', 'Southwest').

    Raises
    ------
    ValueError
        If an unknown resolution specifier is provided.

    Notes
    -----
    The function converts the input angles to complex numbers, computes their mean, and then converts back to an angle.
    If a string resolution is specified, the function maps the average angle to the nearest named direction.

    Examples
    --------
    >>> average_angle([350, 10, 40, 40])
    20.15962133607971
    >>> average_angle([350, 10, 30], resolution=10)
    36.0
    >>> average_angle([350, 10, 30], resolution='four')
    'North'
    >>> average_angle([350, 10, 20], resolution='sixteen')
    'North-northeast'
    """
    complex_angle = np.exp(1j * np.radians(angle))
    angle = wrap_angle(np.degrees(np.angle(complex_angle.mean())))
    if resolution is None:
        return angle

    if not isinstance(resolution, str):
        return wrap_angle(np.round(angle / 360 * resolution) * 360 / resolution)

    resolution = resolution.lower()
    if '4' in resolution or 'four' in resolution:
        resolution = 4
    elif '8' in resolution or 'eight' in resolution:
        resolution = 8
    elif '16' in resolution or 'sixteen' in resolution:
        resolution = 16
    else:
        raise ValueError(f"Unknown resolution specifier '{resolution}'")

    names = [
        (-180., 'south'),
        (-90., 'west'),
        (0., 'north'),
        (90., 'east'),
        (180., 'south'),
    ]

    if resolution >= 8:
        names.extend([
            (-135., 'southwest'),
            (-45., 'northwest'),
            (45., 'northeast'),
            (135., 'southeast'),
        ])
    if resolution >= 16:
        names.extend([
            (-157.5, 'south-southwest'),
            (-112.5, 'west-southwest'),
            (-67.5, 'west-northwest'),
            (-22.5, 'north-northwest'),
            (22.5, 'north-northeast'),
            (67.5, 'east-northeast'),
            (112.5, 'east-southeast'),
            (157.5, 'south-southeast'),
        ])
    name = min([(abs(deg - angle), name) for deg, name in names], key=lambda x: x[0])[1]
    return name.capitalize()


def angle_between(lat, lon, lat_1, lon_1, lat_2, lon_2):
    """Calculate the angle between two coordinates, as seen from a center vertex.

    The angle is counted positive if the second point is clockwise of the first point,
    as seen from the center vertex.

    Parameters
    ----------
    lat : float
        Latitude of the center vertex in degrees.
    lon : float
        Longitude of the center vertex in degrees.
    lat_1 : float
        Latitude of the first point in degrees.
    lon_1 : float
        Longitude of the first point in degrees.
    lat_2 : float
        Latitude of the second point in degrees.
    lon_2 : float
        Longitude of the second point in degrees.

    Returns
    -------
    float
        The angle between the two points as seen from the center vertex, in degrees. The angle is normalized to the range (-180, 180].
    """
    bearing_1 = bearing_to(lat, lon, lat_1, lon_1)
    bearing_2 = bearing_to(lat, lon, lat_2, lon_2)
    return wrap_angle(bearing_2 - bearing_1)



fig = uwacan.visualization.make_chart(center=uwacan.positional.position(latitude_grid.mean(), longitude_grid.mean()), zoom=10)
# fig.add_scattermapbox(lat=[], lon=[])
# fig.add_densitymapbox(
#     lat=np.meshgrid(latitude_grid, longitude_grid, indexing='ij')[0].flatten(),
#     lon=np.meshgrid(latitude_grid, longitude_grid, indexing='ij')[1].flatten(),
#     z=mean_speed.flatten(),
#     radius=5
# )
fig.update_layout(margin={'r': 0.2, 'l': 0.1, 't': 0.1, 'b': 0.1})
colormap = pco.convert_colors_to_same_type(pco.sequential.Viridis, 'rgb')[0].copy()
colormap[0] = 'rgba(' + colormap[0].split('(')[1].split(')')[0] + ', 0.0)'
colormap[1] = 'rgba(' + colormap[1].split('(')[1].split(')')[0] + ', 0.3)'

max_speed = np.max(mean_speed)
max_speed = 12
max_counts = np.max(num_detected)
max_counts = 5000
speed_choices = np.linspace(0, max_speed, 25)
color_choices = pco.sample_colorscale('viridis', speed_choices.size)
def make_color(speed, counts):
    # color = pco.sample_colorscale('viridis', speed / max_speed)[0]
    color = color_choices[np.argmin(np.abs(speed - speed_choices))]
    opacity = (np.clip(counts / max_counts, 0, 1))**0.1
    return 'rgba(' + color.split('(')[1].split(')')[0] + f', {opacity:.6f})'
colors = [
    make_color(s, c)
    for (s, c) in zip(mean_speed.flatten(), num_detected.flatten())
]

fig.add_scattermapbox(
    lat=np.meshgrid(latitude_grid, longitude_grid, indexing='ij')[0].flatten(),
    lon=np.meshgrid(latitude_grid, longitude_grid, indexing='ij')[1].flatten(),
    marker=dict(
        # color=np.where(num_detected > 0, mean_speed, np.nan).flatten(),
        # color=mean_speed.flatten(),
        # cmin=0, cmax=10,
        # colorscale=colormap,
        color = colors,
    ),
    mode='markers',
)
# fig.update_layout(
#   paper_bgcolor='rgba(0,0,0,0)', plot_bgcolor='rgba(0,0,0,0)',
#     # xaxis2=dict(overlaying='x'),
#     # yaxis2=dict(overlaying='y'),
#     xaxis=dict(overlaying='x2'),
#     yaxis=dict(overlaying='y2'),
# )
# colormap = pco.sequential.Viridis

# fig.add_heatmap(
#     x=longitude_grid,
#     y=latitude_grid,
#     z=np.where(num_detected > 0, mean_speed, np.nan),
#     xaxis='x2', yaxis='y2',
#     # opacity=num_detected / np.max(num_detected)
# )
fig.show()


"""Handling of timed sequences, time windows, and extracting time ranges of sequences."""

import datetime
import numpy as np
import abc

# TODO: kill this module and move the parsing and TimeWindow class into the positional module.
# It made more sense to have `Timestamped`, `Referenced`, `Sampled` only for position tracks,
# Everything else is sampled, but is not helped by the `TimedSequence` objects since
# indexing is no longer based on selecting a smaller range of in-memory data.


def parse_timestamp(stamp):
    stamp = ''.join(c for c in stamp if c in '1234567890')
    num_chars = len(stamp)
    year = int(stamp[0:4])
    month = int(stamp[4:6]) if num_chars > 4 else 1
    day = int(stamp[6:8]) if num_chars > 6 else 1
    hour = int(stamp[8:10]) if num_chars > 8 else 0
    minute = int(stamp[10:12]) if num_chars > 10 else 0
    second = int(stamp[12:14]) if num_chars > 12 else 0
    microsecond = int(stamp[14:18].ljust(6, '0')) if num_chars > 12 else 0
    return datetime.datetime(year, month, day, hour, minute, second, microsecond)


class TimedSequence(abc.ABC):
    @property
    @abc.abstractmethod
    def time(self):
        ...

    @property
    def start(self):
        return self._window.start

    @property
    def stop(self):
        return self._window.stop

    @property
    def duration(self):
        return self._window.duration

    @property
    def center(self):
        return self._window.center

    @abc.abstractmethod
    def time_range(self, start, stop):
        if isinstance(start, datetime.datetime):
            start = (start - self.start).total_seconds()

        if isinstance(stop, datetime.datetime):
            stop = (stop - self.start).total_seconds()

        if start < 0:
            raise ValueError('Cannot access signal before measurement start')
        if start > self.duration:
            raise ValueError('Cannot access signal after measurement stop')
        if stop < 0:
            raise ValueError('Cannot access signal before measurement start')
        if stop > self.duration:
            raise ValueError('Cannot access signal after measurement stop')

        return start, stop


class TimedstampedSequence(TimedSequence):
    def __init__(self, timestamps):
        self.timestamps = timestamps
        self._window = TimeWindow(start=timestamps[0], stop=timestamps[-1])

    @property
    def time(self):
        np.asarray([
            (stamp - self.start).total_seconds()
            for stamp in self.timestamps
        ])

    def time_range(self, start, stop):
        start, stop = super().time_range(start, stop)
        for idx, stamp in enumerate(self.timestamps):
            if (stamp - self.start).total_seconds() >= start:
                start_idx = idx
                break

        for idx, stamp in enumerate(self.timestamps[start_idx:], start_idx):
            if (stamp - self.start).total_seconds() >= stop:
                stop_idx = idx
                break

        return start_idx, stop_idx


class ReferencedSequence(TimedSequence):
    def __init__(self, times, reference):
        self._times = times
        self._reference = reference
        self._window = TimeWindow(
            start=reference + datetime.timedelta(seconds=times[0]),
            stop=reference + datetime.timedelta(seconds=times[-1]),
        )

    @property
    def time(self):
        return self._times - (self.start - self._reference).total_seconds()

    def time_range(self, start, stop):
        start, stop = super().time_range(start, stop)

        start_idx = np.searchsorted(self._times, start)
        stop_idx = np.searchsorted(self._times, stop)
        return start_idx, stop_idx


class SampledSequence(TimedSequence):
    def __init__(self, samplerate, start, num_samples):
        self.samplerate = samplerate
        self.num_samples = num_samples
        self._start = start
        self._window = TimeWindow(start=start, duration=num_samples / samplerate)

    @property
    def time(self):
        return np.arange(self.num_samples) / self.samplerate

    def time_range(self, start, stop):
        start, stop = super().time_range(start, stop)
        start_idx = round(start * self.samplerate)
        stop_idx = round(stop * self.samplerate)
        return start_idx, stop_idx


class TimeWindow:
    def __init__(self, start=None, stop=None, center=None, duration=None):
        if isinstance(start, str):
            start = parse_timestamp(start)
        if isinstance(stop, str):
            stop = parse_timestamp(stop)
        if isinstance(center, str):
            center = parse_timestamp(center)

        if None not in (start, stop):
            self.start = start
            self.stop = stop
            start = stop = None
        elif None not in (center, duration):
            self.start = center - datetime.timedelta(seconds=duration / 2)
            self.stop = center + datetime.timedelta(seconds=duration / 2)
            center = duration = None
        elif None not in (start, duration):
            self.start = start
            self.stop = start + datetime.timedelta(seconds=duration)
            start = duration = None
        elif None not in (stop, duration):
            self.stop = stop
            self.start = stop - datetime.timedelta(seconds=duration)
            stop = duration = None
        elif None not in (start, center):
            self.start = start
            self.stop = start + (center - start) / 2
            start = center = None
        elif None not in (stop, center):
            self.stop = stop
            self.start = stop - (stop - center) / 2
            stop = center = None

        if (start, stop, center, duration) != (None, None, None, None):
            raise ValueError('Cannot input more than two input arguments to a time window!')

    @property
    def duration(self):
        return (self.stop - self.start).total_seconds()

    @property
    def center(self):
        return self.start + datetime.timedelta(seconds=self.duration / 2)



"""Implementations for spectral filterbanks.

.. currentmodule:: uwacan._filterbank

Core processing and analysis
----------------------------
.. autosummary::
    :toctree: generated

    Filterbank
    FilterbankRoller
    spectrum
    linear_to_banded

"""


from . import _core
import xarray as xr
import numpy as np
import scipy.signal
import numba


def spectrum(time_data, window=None, scaling="density", nfft=None, detrend=True, samplerate=None, axis=None):
    """Compute the power spectrum of time-domain data.

    The `spectrum` function calculates the power spectrum of input time-series data. It supports
    various input types, including `~uwacan.TimeData`, `xarray.DataArray`, and NumPy arrays. The
    function applies windowing, detrending, and scaling as specified by the parameters to
    produce the frequency-domain representation of the data.

    Parameters
    ----------
    time_data : _core.TimeData or xr.DataArray or numpy.ndarray
        The input time-domain data to compute the spectrum for. The data can be one of the following:

        - `~uwacan.TimeData`: Wrapped time data from ``uwacan``.
        - `xarray.DataArray`: An xarray DataArray with a 'time' dimension.
        - `numpy.ndarray`: A NumPy array containing time-series data.

    window : str or array_like, optional
        The window function to apply to the data before computing the FFT. This can be:

        - A string specifying the type of window to use (e.g., ``"hann"``, ``"kaiser"``, ``"blackman"``).
        - An array-like sequence of window coefficients.
        - If ``None``, no window is applied. Default is ``None``.

    scaling : {'density', 'spectrum', 'dc-nyquist'} or numeric, optional
        Specifies the scaling of the power spectrum. Options include:

        - ``'density'``: Computes the power spectral density.
        - ``'spectrum'``: Computes the power spectrum.
        - ``'dc-nyquist'``: Halves the output at DC and Nyquist frequencies. Use with a pre-scaled window that takes care to scale the remainder of the single-sided spectrum.
        - any numeric value: The output of the fft will be scaled by this value.

        Default is ``"density"``.

    nfft : int, optional
        The number of points to use in the FFT computation. If ``None``, it defaults to the length
        of the input data along the specified axis.

    detrend : bool, default=True
        If ``True``, removes the mean from the data before computing the FFT to reduce spectral leakage.
        If ``False``, no detrending is performed.

    samplerate : float, optional
        The sampling rate of the input data in Hz. Required if ``time_data`` is an numpy array.
        If not provided, it defaults to 1. This parameter is used to compute the frequency axis and proper density scaling.

    axis : int, optional
        The axis along which to compute the FFT. If ``None``, the last axis is used.
        Only used for numpy inputs.
        This parameter allows flexibility in handling multi-dimensional data.

    Returns
    -------
    _core.FrequencyData or xr.DataArray or numpy.ndarray
        The computed power spectrum of the input data. The return type matches the input type:

        - If ``time_data`` is a `~uwacan.TimeData`, returns a `~uwacan.FrequencyData` object.
        - If ``time_data`` is an `xarray.DataArray`, returns an `xarray.DataArray` with a 'frequency' dimension.
        - If ``time_data`` is a `numpy.ndarray`, returns a NumPy array containing the power spectrum.

    """
    if isinstance(time_data, _core.TimeData):
        return _core.FrequencyData(spectrum(time_data.data, window=window, scaling=scaling, nfft=nfft, detrend=detrend))
    if isinstance(time_data, xr.DataArray):
        freq_data = xr.apply_ufunc(
            spectrum,
            time_data,
            input_core_dims=[["time"]],
            output_core_dims=[["frequency"]],
            kwargs=dict(
                window=window,
                scaling=scaling,
                nfft=nfft,
                detrend=detrend,
                samplerate=samplerate or time_data.time.attrs.get("rate", None),
            ),
        )
        freq_data.coords["frequency"] = np.fft.rfftfreq(nfft or time_data.time.size, 1 / time_data.time.rate)
        freq_data.coords["time"] = time_data.time[0] + np.timedelta64(
            round(time_data.time.size / 2 / time_data.time.rate * 1e9), "ns"
        )
        return freq_data

    if axis is not None:
        time_data = np.moveaxis(time_data, axis, -1)

    if detrend:
        time_data = time_data - time_data.mean(axis=-1, keepdims=True)

    if window is not None:
        if not isinstance(window, np.ndarray):
            window = scipy.signal.windows.get_window(window, time_data.shape[-1], False)
        time_data = time_data * window

    nfft = nfft or time_data.shape[-1]
    freq_data = np.fft.rfft(time_data, axis=-1, n=nfft)
    freq_data = np.abs(freq_data) ** 2

    if scaling == "density":
        samplerate = samplerate or 1
        if window is not None:
            scaling = 2 / (np.sum(window**2) * samplerate)
        else:
            scaling = 2 / (time_data.shape[-1] * samplerate)
    elif scaling == "spectrum":
        if window is not None:
            scaling = 2 / np.sum(window) ** 2
        else:
            scaling = 2 / time_data.shape[-1]
    elif scaling == "dc-nyquist":
        # Remove doubling of DC
        freq_data[..., 0] /= 2
        if nfft % 2 == 0:
            # Even size, remove doubling of Nyquist
            freq_data[..., -1] /= 2
        scaling = False

    if scaling:
        freq_data *= scaling

        # Remove doubling of DC
        freq_data[..., 0] /= 2
        if nfft % 2 == 0:
            # Even size, remove doubling of Nyquist
            freq_data[..., -1] /= 2

    if axis is not None:
        freq_data = np.moveaxis(freq_data, -1, axis)

    return freq_data


@numba.njit()
def _linear_to_banded(linear_spectrum, lower_edges, upper_edges, spectral_resolution):
    banded_spectrum = np.full(lower_edges.shape + linear_spectrum.shape[1:], np.nan)
    for band_idx, (lower_edge, upper_edge) in enumerate(zip(lower_edges, upper_edges)):
        lower_idx = int(np.floor(lower_edge / spectral_resolution + 0.5))  # (l_idx - 0.5) * Δf = l
        upper_idx = int(np.ceil(upper_edge / spectral_resolution - 0.5))  # (u_idx + 0.5) * Δf = u
        lower_idx = max(lower_idx, 0)
        upper_idx = min(upper_idx, linear_spectrum.shape[0] - 1)

        if lower_idx == upper_idx:
            # This can only happen if both frequencies l and u are within the same fft bin.
            # Since we don't allow the fft bins to be larger than the output bins, we thus have the exact same band.
            banded_spectrum[band_idx] = linear_spectrum[lower_idx]
        else:
            # weight edge bins by "(whole bin - what is not in the new band) / whole bin"
            # lower fft bin edge l_e = (l_idx - 0.5) * Δf
            # w_l = (Δf - (l - l_e)) / Δf = l_idx + 0.5 - l / Δf
            first_weight = lower_idx + 0.5 - lower_edge / spectral_resolution
            # upper fft bin edge u_e = (u_idx + 0.5) * Δf
            # w_u = (Δf - (u_e - u)) / Δf = 0.5 - u_idx + u / Δf
            last_weight = upper_edge / spectral_resolution - upper_idx + 0.5
            # Sum the components fully within the output bin `[l_idx + 1:u_idx]`, and weighted components partially in the band.
            this_band = (
                linear_spectrum[lower_idx + 1 : upper_idx].sum(axis=0)
                + linear_spectrum[lower_idx] * first_weight
                + linear_spectrum[upper_idx] * last_weight
            )
            banded_spectrum[band_idx] = this_band * (
                spectral_resolution / (upper_edge - lower_edge)
            )  # Rescale the power density.
    return banded_spectrum


def linear_to_banded(linear_spectrum, lower_edges, upper_edges, spectral_resolution, axis=0):
    """Aggregate a linear power spectrum into specified frequency bands.

    The `linear_to_banded` function converts a linear power spectrum into a banded spectrum by
    summing power within frequency bands defined by ``lower_edges`` and ``upper_edges``. It handles
    multi-dimensional spectra by allowing specification of the axis corresponding to frequency
    bins.

    Parameters
    ----------
    linear_spectrum : `numpy.ndarray`
        The input linear power spectrum. The axis specified by ``axis`` should correspond to
        frequency bins.
    lower_edges : array_like
        The lower frequency edges for each band. Must be in ascending order.
    upper_edges : array_like
        The upper frequency edges for each band. Must be in ascending order and greater
        than or equal to ``lower_edges``.
    spectral_resolution : float
        The frequency resolution (Δf) of the linear spectrum.
    axis : int, optional, default=0
        The axis of ``linear_spectrum`` that corresponds to frequency bins. If the frequency
        bins are not along the first axis, specify the appropriate axis index.

    Returns
    -------
    banded_spectrum : numpy.ndarray
        The aggregated banded power spectrum.

    """
    # TODO: add features here to allow non-numpy inputs. Simply unwrap and rewrap as needed.
    if axis:
        linear_spectrum = np.moveaxis(linear_spectrum, axis, 0)
    banded = _linear_to_banded(linear_spectrum, lower_edges, upper_edges, spectral_resolution)
    if axis:
        banded = np.moveaxis(banded, 0, axis)
    return banded


class FilterbankRoller(_core.Roller):
    """Rolling computation of power spectra and spectrograms, both linear and banded.

    Parameters
    ----------
    time_data : TimeData
        The time-data wrapper to process.
    duration : float, optional
        The duration of the fft windows, in seconds.
    step : float, optional
        The step size between consecutive fft windows, in seconds.
    overlap : float, optional
        The amount of overlap between fft windows, as a fraction of the duration.
    min_frequency : float
        The lowest frequency to include in the processing.
    max_frequency : float
        The highest frequency to include in the processing.
    bands_per_decade : float, optional
        The number of frequency bands per decade for logarithmic scaling.
    hybrid_resolution : float
        A frequency resolution to aim for. Only used if ``frame_duration`` is not given
    fft_window : str, default="hann"
        The window function to apply to each rolling window before computing the FFT.
        Can be a string specifying a window type (e.g., ``"hann"``, ``"kaiser"``, ``"blackman"``)
        or an array-like sequence of window coefficients..

    Notes
    -----
    The processing is done in stft frames determined by ``duration``, ``step``
    ``overlap``, and ``hybrid_resolution``. At least one of ``duration``, ``step``,
    or ``resolution`` has to be given, see `~_core.time_frame_settings` for further details.
    At least one of ``min_frequency`` and ``hybrid_resolution`` has to be given.
    Note that the ``duration`` and ``step`` can be auto-chosen from the overlap
    and required frequency resolution, either from ``hybrid_resolution`` or ``min_frequency``.

    Raises
    ------
    ValueError
        If the processing settings are not compatible, e.g.,
        - frequency bands with bandwidth smaller than the frame duration allows
    """

    def __init__(
        self,
        time_data,
        duration=None,
        step=None,
        overlap=None,
        min_frequency=None,
        max_frequency=None,
        bands_per_decade=None,
        hybrid_resolution=None,
        fft_window="hann",
    ):
        self.settings = _core.time_frame_settings(
            duration=duration,
            step=step,
            overlap=overlap,
            resolution=None if isinstance(hybrid_resolution, bool) else hybrid_resolution,
            signal_length=time_data.time_window.duration,
            samplerate=time_data.samplerate,
        )
        self.min_frequency = min_frequency or 0
        self.max_frequency = max_frequency or time_data.samplerate / 2

        self.bands_per_decade = bands_per_decade
        self.hybrid_resolution = hybrid_resolution
        self.fft_window = fft_window

        self.time_data = time_data
        self.roller = self.time_data.rolling(
            duration=self.settings["duration"], step=self.settings["step"], overlap=self.settings["overlap"]
        )

        self.processing_axis = self.roller.dims.index("time")
        self.check_frequency_resolution()
        self.make_frequency_vectors()

    @property
    def dims(self):  # noqa: D102
        dims = list(self.roller.dims)
        dims[self.processing_axis] = "frequency"
        return tuple(dims)

    @property
    def shape(self):  # noqa: D102
        shape = list(self.roller.shape)
        shape[self.processing_axis] = len(self.frequency)
        return tuple(shape)

    @property
    def coords(self):  # noqa: D102
        coords = dict(self.roller.coords)
        del coords["time"]
        coords["frequency"] = xr.DataArray(self.frequency, dims="frequency", coords={"frequency": self.frequency})
        return coords

    def check_frequency_resolution(self):
        """Validate the frequency resolution against the temporal window settings."""
        # TODO: Check what this used to do when you used parameters from a spectrogram object.
        if not self.bands_per_decade:
            self.bands_per_decade = False
            self.hybrid_resolution = False
        else:
            if not self.hybrid_resolution:
                self.hybrid_resolution = False
                # Not using hybrid, we need long enough frames to compute the lowest band.
                lowest_bandwidth = self.min_frequency * (
                    10 ** (0.5 / self.bands_per_decade) - 10 ** (-0.5 / self.bands_per_decade)
                )
                if lowest_bandwidth * self.settings["duration"] < 1:
                    raise ValueError(
                        f'{self.bands_per_decade}th-decade filter band at {self.min_frequency:.2f} Hz with bandwidth of {lowest_bandwidth:.2f} Hz '
                        f'cannot be calculated from temporal windows of {self.settings["duration"]:.2f} s.'
                    )
            else:
                # Get the hybrid resolution settings.
                if self.hybrid_resolution is True:
                    self.hybrid_resolution = 1 / self.settings["duration"]
                if self.hybrid_resolution * self.settings["duration"] < 1:
                    raise ValueError(
                        f'Hybrid filterbank with resolution of {self.hybrid_resolution:.2f} Hz '
                        f'cannot be calculated from temporal windows of {self.settings["duration"]:.2f} s.'
                    )

    def make_frequency_vectors(self):
        """Construct frequency vectors and band definitions based on spectrogram settings."""
        nfft = self.settings["samples_per_frame"]
        self.linear_frequency = np.arange(nfft // 2 + 1) * self.time_data.samplerate / nfft
        self.bandwidth = self.linear_frequency[1]

        if self.max_frequency < self.linear_frequency[-1]:
            upper_index = int(np.floor(self.max_frequency / self.time_data.samplerate * nfft))
        else:
            upper_index = None

        if self.min_frequency > 0:
            lower_index = int(np.ceil(self.min_frequency / self.time_data.samplerate * nfft))
        else:
            lower_index = None

        if upper_index or lower_index:
            self.linear_slice = (slice(None),) * self.processing_axis + (slice(lower_index, upper_index),)
        else:
            self.linear_slice = False

        if self.linear_slice:
            self.frequency = self.linear_frequency[self.linear_slice]

        if self.bands_per_decade:
            log_band_scaling = 10 ** (0.5 / self.bands_per_decade)
            if self.hybrid_resolution:
                # The frequency at which the logspaced bands cover at least one linspaced band
                minimum_bandwidth_frequency = self.hybrid_resolution / (log_band_scaling - 1 / log_band_scaling)
                first_log_idx = int(
                    np.ceil(self.bands_per_decade * np.log10(minimum_bandwidth_frequency / 1e3))
                )
                last_linear_idx = int(np.floor(minimum_bandwidth_frequency / self.hybrid_resolution))

                # Since the logspaced bands have pre-determined centers, we can't just start them after the linspaced bands.
                # Often, the bands will overlap at the minimum bandwidth frequency, so we look for the first band
                # that does not overlap, i.e., the upper edge of the last linspaced band is below the lower edge of the first
                # logspaced band
                while (last_linear_idx + 0.5) * self.hybrid_resolution > 1e3 * 10 ** (
                    (first_log_idx - 0.5) / self.bands_per_decade
                ):
                    # Condition is "upper edge of last linear band is higher than lower edge of first logarithmic band"
                    last_linear_idx += 1
                    first_log_idx += 1

                if last_linear_idx * self.hybrid_resolution > self.max_frequency:
                    last_linear_idx = int(np.floor(self.max_frequency / self.hybrid_resolution))
                first_linear_idx = int(np.ceil(self.min_frequency / self.hybrid_resolution))
            else:
                first_linear_idx = last_linear_idx = 0
                first_log_idx = np.round(self.bands_per_decade * np.log10(self.min_frequency / 1e3))

            last_log_idx = round(self.bands_per_decade * np.log10(self.max_frequency / 1e3))

            lin_centers = np.arange(first_linear_idx, last_linear_idx) * self.hybrid_resolution
            lin_lowers = lin_centers - 0.5 * self.hybrid_resolution
            lin_uppers = lin_centers + 0.5 * self.hybrid_resolution

            log_centers = 1e3 * 10 ** (np.arange(first_log_idx, last_log_idx + 1) / self.bands_per_decade)
            log_lowers = log_centers / log_band_scaling
            log_uppers = log_centers * log_band_scaling

            centers = np.concatenate([lin_centers, log_centers])
            lowers = np.concatenate([lin_lowers, log_lowers])
            uppers = np.concatenate([lin_uppers, log_uppers])

            if centers[0] < self.min_frequency:
                mask = centers >= self.min_frequency
                lowers = lowers[mask]
                centers = centers[mask]
                uppers = uppers[mask]
            if centers[-1] > self.max_frequency:
                mask = centers <= self.max_frequency
                lowers = lowers[mask]
                centers = centers[mask]
                uppers = uppers[mask]

            self.band_lower_edges = lowers
            self.band_centers = centers
            self.band_upper_edges = uppers
            self.frequency = centers
            self.bandwidth = uppers - lowers

    def numpy_frames(self):  # noqa: D102
        window = scipy.signal.windows.get_window(self.fft_window, self.settings["samples_per_frame"], False)
        window /= ((window**2).sum() * self.time_data.samplerate / 2) ** 0.5

        for idx, time_frame in enumerate(self.roller.numpy_frames()):
            freq_frame = spectrum(time_frame, window=window, scaling="dc-nyquist", axis=self.processing_axis)
            if self.bands_per_decade:
                freq_frame = linear_to_banded(
                    freq_frame,
                    lower_edges=self.band_lower_edges,
                    upper_edges=self.band_upper_edges,
                    spectral_resolution=self.settings["resolution"],
                    axis=self.processing_axis,
                )
            elif self.linear_slice:
                freq_frame = freq_frame[self.linear_slice]
            yield freq_frame

    def __iter__(self):
        start_time = _core.time_to_np(self.time_data.time_window.start)
        start_time += np.timedelta64(
            int(self.settings["samples_per_frame"] / 2 / self.time_data.samplerate * 1e9), "ns"
        )
        for frame_idx, frame in enumerate(self.numpy_frames()):
            time_since_start = frame_idx * self.settings["sample_step"] / self.time_data.samplerate
            time_since_start = np.timedelta64(int(time_since_start * 1e9), "ns")
            frame = _core.FrequencyData(
                frame,
                frequency=self.frequency,
                bandwidth=self.bandwidth,
                coords=self.coords,
                dims=self.dims,
            )
            frame.data["time"] = start_time + time_since_start
            yield frame

    def batches(self, batch_size):
        """Generate batches of spectrogram frames.

        Parameters
        ----------
        batch_size : int
            Number of frames to include in each batch.

        Yields
        ------
        TimeFrequencyData
            Batches of spectrogram frames with time and frequency information.
            Each batch contains up to `batch_size` frames, with the final batch
            potentially containing fewer frames if the total number of frames
            is not evenly divisible by the batch size.
        """
        batch_output = np.zeros((batch_size,) + self.shape)
        for frame_idx, frame in enumerate(self.numpy_frames()):
            if frame_idx % batch_size == 0:
                batch_start_time = self.time_data.time_window.start.add(seconds=frame_idx * self.settings["step"])
            batch_output[frame_idx % batch_size] = frame
            if (frame_idx + 1) % batch_size == 0 or frame_idx == self.num_frames - 1:
                batch_data = _core.TimeFrequencyData(
                    batch_output[:frame_idx % batch_size + 1].copy(),  # Only use the frames we've filled
                    frequency=self.frequency,
                    bandwidth=self.bandwidth,
                    samplerate=self.time_data.samplerate / self.settings["sample_step"],
                    start_time=batch_start_time.add(seconds=self.settings["duration"] / 2),
                    coords=self.coords,
                    dims=("time",) + self.dims,
                    attrs=dict(
                        frame_duration=self.settings["duration"],
                        frame_overlap=self.settings["overlap"],
                        frame_step=self.settings["step"],
                        bands_per_decade=self.bands_per_decade,
                        hybrid_resolution=self.hybrid_resolution,
                    ),
                )
                yield batch_data


class Filterbank:
    """Calculates spectrograms, both linear and banded.

    Parameters
    ----------
    bands_per_decade : float, optional
        The number of frequency bands per decade for logarithmic scaling.
    frame_step : float
        The time step between stft frames, in seconds.
    frame_duration : float
        The duration of each stft frame, in seconds.
    frame_overlap : float, default=0.5
        The overlap factor between stft frames. A negative value leaves
        gaps between frames.
    min_frequency : float
        The lowest frequency to include in the processing.
    max_frequency : float
        The highest frequency to include in the processing.
    hybrid_resolution : float
        A frequency resolution to aim for. Only used if ``frame_duration`` is not given
    fft_window : str, default="hann"
        The window function to apply to each rolling window before computing the FFT.
        Can be a string specifying a window type (e.g., ``"hann"``, ``"kaiser"``, ``"blackman"``)
        or an array-like sequence of window coefficients..

    Notes
    -----
    The processing is done in stft frames determined by ``frame_duration``, ``frame_step``
    ``frame_overlap``, and ``hybrid_resolution``. At least one of ``duration``, ``step``,
    or ``resolution`` has to be given, see `~_core.time_frame_settings` for further details.
    At least one of ``min_frequency`` and ``hybrid_resolution`` has to be given.
    Note that the ``frame_duration`` and ``frame_step`` can be auto-chosen from the overlap
    and required frequency resolution, either from ``hybrid_resolution`` or ``min_frequency``.

    Raises
    ------
    ValueError
        If the processing settings are not compatible, e.g.,
        - frequency bands with bandwidth smaller than the frame duration allows
    """

    def __init__(
        self,
        *,
        bands_per_decade=None,
        frame_step=None,
        frame_duration=None,
        frame_overlap=0.5,
        min_frequency=None,
        max_frequency=None,
        hybrid_resolution=None,
        fft_window="hann",
    ):
        self.frame_duration = frame_duration
        self.frame_overlap = frame_overlap
        self.frame_step = frame_step
        self.fft_window = fft_window
        self.bands_per_decade = bands_per_decade
        self.min_frequency = min_frequency
        self.max_frequency = max_frequency
        self.hybrid_resolution = hybrid_resolution

    def rolling(self, time_data):
        roller = FilterbankRoller(
            time_data=time_data,
            duration=self.frame_duration,
            step=self.frame_step,
            overlap=self.frame_overlap,
            min_frequency=self.min_frequency,
            max_frequency=self.max_frequency,
            bands_per_decade=self.bands_per_decade,
            hybrid_resolution=self.hybrid_resolution,
            fft_window=self.fft_window,
        )
        return roller

    def __call__(self, time_data):
        """Process time data to spectrograms.

        Parameters
        ----------
        time_data : `~uwacan.TimeData` or `~uwacan.recordings.AudioFileRecording`
            The data to process.

        Returns
        -------
        filtered_data : `~uwacan.TimeFrequencyData`
        """
        roller = self.rolling(time_data)
        return next(roller.batches(roller.num_frames))



import numpy as np
import scipy.signal
from . import positional, _core
import itertools
import abc



class DataTree(_core.Branch):
    def __new__(cls, *args, children=None, **kwargs):
        if children is not None:
            first = next(iter(children.values()))
            obj = super().__new__(first._leaf_type._tree_class)
        else:
            obj = super().__new__(cls)
        if not isinstance(obj, cls):
            obj.__init__(*args, children=children, **kwargs)
        return obj

    @property
    def data(self):
        return np.stack([child.data for child in self.values()], axis=0)


class TimeDataTree(_core.TimeBranchin, DataTree):
    ...


class FrequencyDataTree(DataTree):
    @property
    def frequency(self):
        leaves = self._traverse(leaves=True, branches=False, root=False)
        frequency = next(leaves).frequency
        for leaf in leaves:
            leaf_frequency = leaf.frequency
            if not (frequency is leaf_frequency or (
                np.shape(frequency) == np.shape(leaf_frequency)
                and np.allclose(frequency, leaf_frequency)
            )):
                raise ValueError('Cannot access frequency of DataTree where the frequencies of the leaves is not consistent.')
        return frequency


class TimeFrequencyDataTree(TimeDataTree, FrequencyDataTree):
    ...


class Data(_core.Leaf):
    dims = tuple()
    _tree_class = DataTree

    def __new__(cls, data, dims=None, **kwargs):
        if dims is not None:
            if 'time' in dims and 'frequency' in dims:
                obj = super().__new__(TimeFrequency)
            elif 'time' in dims:
                obj = super().__new__(Time)
            elif 'frequency' in dims:
                obj = super().__new__(Frequency)
            else:
                obj = super().__new__(Data)
        else:
            obj = super().__new__(cls)
        if not isinstance(obj, cls):
            obj.__init__(data, dims=dims, **kwargs)
        return obj

    def __init__(self, data, dims=None, **kwargs):
        super().__init__(**kwargs)
        self._data = np.asarray(data)

        if dims is not None:
            self.dims = dims
        if len(self.dims) != self.data.ndim:
            raise ValueError('The number of dimensions in the data does not match the number of expected axes')

    def clone(self, data, dims=None):
        return type(self)(data, dims=dims or self.dims, metadata=self.metadata)

    def reduce(self, function, dim, *args, **kwargs):
        if isinstance(dim, (int, str)):
            dim = [dim]
        reduce_axes = []
        for ax in dim:
            if isinstance(ax, str):
                ax = self.dims.index(ax)
            reduce_axes.append(ax)
        reduce_axes = tuple(reduce_axes)
        new_dims = tuple(dim for idx, dim in enumerate(self.dims) if idx not in reduce_axes)

        if isinstance(function, _core.Reduction):
            function = function.function

        out = function(self.data, axis=reduce_axes, *args, **kwargs)
        if not isinstance(out, Data):
            out = self.clone(out, dims=new_dims)
        return out


class Time(_core.TimeLeafin, Data):
    dims = ('time',)
    _tree_class = TimeDataTree

    def __init__(self, data, samplerate, start_time, **kwargs):
        super().__init__(data=data, **kwargs)
        self.samplerate = samplerate
        # TODO: parse the start time here as well, if it's a string.
        self._start_time = start_time

    def clone(self, data, dims=None):
        dims = self.dims if dims is None else dims
        kwargs = {'dims': dims, 'metadata': self.metadata}
        if 'time' in dims:
            kwargs.update(samplerate=self.samplerate, start_time=self._start_time)
        return type(self)(data, **kwargs)

    @property
    def num_samples(self):
        return self.data.shape[-1]

    @property
    def relative_time(self):
        return np.arange(self.num_samples) / self.datarate

    @property
    def timestamps(self):
        return [self._start_time + positional.datetime.timedelta(seconds=t) for t in self.relative_time]

    @property
    def time_period(self):
        return _core.TimePeriod(start=self._start_time, duration=self.data.shape[-1] / self.samplerate)

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        window = self.time_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        start = (window.start - self._start_time).total_seconds()
        stop = (window.stop - self._start_time).total_seconds()
        # Indices assumed to be seconds from start
        start = np.math.floor(start * self.datarate)
        stop = np.math.ceil(stop * self.datarate)
        return type(self)(
            data=self.data[..., start:stop],
            start_time=self._start_time + positional.datetime.timedelta(seconds=start / self.samplerate),
            samplerate=self.samplerate,
            metadata=self.metadata,
        )


class Pressure(Time):
    @classmethod
    def from_raw_and_calibration(cls, data, calibration, **kwargs):
        """Create a pressure signal from raw values and known calibration.

        Parameters
        ----------
        data : ndarray
            The raw unscaled input data.
            Should be shape `(n_ch, n_samp)` or 1d with just `n_samp`
        calibration : numeric
            The calibration given as dB re. U/μPa,
            where U is the units of the `data`.
            If `data` is in volts, give the calibration in
            dB re. 1V/μPa.
            If `data` is in "file units", e.g. scaled to [-1, 1] fullscale,
            the calibration value must scale from "file units" to μPa.
        """
        data = np.asarray(data)
        calibration = np.asarray(calibration).astype('float32')
        c = 10**(calibration / 20) / 1e-6  # Calibration values are given as dB re. 1μPa
        c = c.reshape((-1,) + (1,) * (data.ndim - 1))
        if data.dtype in (np.int8, np.int16, np.int32, np.float32):
            c = c.astype(np.float32)

        calibrated = data / c
        obj = cls(data=calibrated, **kwargs)
        return obj


class Frequency(Data):
    dims = ('frequency',)
    _tree_class = FrequencyDataTree

    def __init__(self, data, frequency, bandwidth, **kwargs):
        super().__init__(data=data, **kwargs)
        self.frequency = frequency
        self.bandwidth = bandwidth

    def clone(self, data, dims=None):
        dims = dims or self.dims
        kwargs = {'dims': dims, 'metadata': self.metadata}
        if 'frequency' in dims:
            kwargs.update(frequency=self.frequency, bandwidth=self.bandwidth)
        return type(self)(data, **kwargs)


class TimeFrequency(Time, Frequency):
    dims = ('frequency', 'time')
    _tree_class = TimeFrequencyDataTree

    def clone(self, data, dims=None):
        dims = self.dims if dims is None else dims
        kwargs = {'dims': dims, 'metadata': self.metadata}
        if 'time' in dims:
            kwargs.update(samplerate=self.samplerate, start_time=self._start_time)
        if 'frequency' in dims:
            kwargs.update(frequency=self.frequency, bandwidth=self.bandwidth)
        return type(self)(data, **kwargs)


class DataStack(_core.Branch):
    @property
    def data(self):
        return np.stack([child.data for child in self._children], axis=0)



class ClassificationLimit:
    def __init__(self, *limits, is_density, is_monopole):
        self.limits = limits
        self.is_density = is_density
        self.is_monopole = is_monopole

    def __call__(self, frequency, density=True, monopole=False):
        frequency = np.asarray(frequency).astype(float)
        conditions = [
            (limit.f_low <= frequency) & (frequency <= limit.f_high)
            for limit in self.limits
        ]
        limits = [limit.limit for limit in self.limits] + [np.nan]
        levels = np.piecewise(frequency, conditions, limits)

        # TODO: this only works if everything is in thirdoctavebands
        if self.is_density and not density:
            levels += 10 * np.log10((2**(1/6) - 2**(-1/6)) * frequency)
        elif not density:
            levels -= 10 * np.log10((2**(1/6) - 2**(-1/6)) * frequency)

        return levels

    def lloyd_mirror_compensation(self, frequency, depth, angle):
        denom = 4 * np.pi * frequency * depth + np.sin(np.radians(angle) / 1500)
        return np.maximum(0, 10 * np.log10(0.5 + 1 / denom**2))

bureau_veritas_controlled = ClassificationLimit(
    limit(10, 50, lambda f: 169 - 2 * np.log10(f)),
    limit(50, 1e3, lambda f: 165.6 - 20 * np.log10(f / 50)),
    limit(1e3, 50e3, lambda f: 139.6 - 20 * np.log10(f / 1000)),
    is_density=True, is_monopole=False,
)

bureau_veritas_controlled = ClassificationLimit(
    limit(10, 50, lambda f: 169 - 2 * np.log10(f)),
    limit(50, 1e3, lambda f: 165.6 - 20 * np.log10(f / 50)),
    limit(1e3, 50e3, lambda f: 139.6 - 20 * np.log10(f / 1000)),
    is_density=True, is_monopole=False,
)



# class _SampleTimer:
#     def __init__(self, xr_obj):
#         if 'time' not in xr_obj.dims:
#             raise TypeError(".sampling accessor only available for xarrays with 'time' coordinate")
#         self._xr_obj = xr_obj

#     @property
#     def rate(self):
#         return self._xr_obj.coords['time'].attrs['rate']

#     @property
#     def num(self):
#         return self._xr_obj.sizes['time']

#     @property
#     def window(self):
#         start = positional.time_to_datetime(self._xr_obj.time.data[0])
#         # Calculating duration from number and rate means the stop points to the sample after the last,
#         # which is more intuitive when considering signal durations etc.
#         return positional.TimeWindow(
#             start=start,
#             duration=self.num / self.rate,
#         )

#     def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
#         original_window = self.window
#         new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
#         if isinstance(new_window, positional.TimeWindow):
#             start = (new_window.start - original_window.start).total_seconds()
#             stop = (new_window.stop - original_window.start).total_seconds()
#             # Indices assumed to be seconds from start
#             start = np.math.floor(start * self.rate)
#             stop = np.math.ceil(stop * self.rate)
#             idx = slice(start, stop)
#         else:
#             idx = (new_window - original_window.start).total_seconds()
#             idx = round(idx * self.rate)

#         new_obj = self._xr_obj.isel(time=idx)
#         return new_obj


# class _StampedTimer:
#     def __init__(self, xr_obj):
#         self._xr_obj = xr_obj

#     @property
#     def window(self):
#         start = positional.time_to_datetime(self._xr_obj.time.data[0])
#         stop = positional.time_to_datetime(self._xr_obj.time.data[-1])
#         return positional.TimeWindow(start=start, stop=stop)

#     def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
#         original_window = self.window
#         new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
#         if isinstance(new_window, positional.TimeWindow):
#             start = new_window.start.in_tz('UTC').naive()
#             stop = new_window.stop.in_tz('UTC').naive()
#             return self._xr_obj.sel(time=slice(start, stop))
#         else:
#             return self._xr_obj.sel(time=new_window.in_tz('UTC').naive(), method='nearest')


# def _make_sampler(xr_obj):
#     if 'time' not in xr_obj.dims:
#         raise TypeError(".sampling accessor only available for xarrays with 'time' coordinate")
#     if 'rate' in xr_obj.time.attrs:
#         return _SampleTimer(xr_obj)
#     else:
#         return _StampedTimer(xr_obj)


# xr.register_dataarray_accessor('sampling')(_make_sampler)
# xr.register_dataset_accessor('sampling')(_make_sampler)

# def time_data(data, start_time, samplerate, dims=None, coords=None):
#     if not isinstance(data, xr.DataArray):
#         if dims is None:
#             if data.ndim == 1:
#                 dims = 'time'
#             else:
#                 raise ValueError(f'Cannot guess dimensions for time data with {data.ndim} dimensions')
#         data = xr.DataArray(data, dims=dims)

#     n_samples = data.sizes['time']
#     start_time = positional.time_to_np(start_time)
#     offsets = np.arange(n_samples) * 1e9 / samplerate
#     time = start_time + offsets.astype('timedelta64[ns]')
#     data = data.assign_coords(
#         time=('time', time, {'rate': samplerate}),
#         **{name: coord for (name, coord) in (coords or {}).items() if name != 'time'}
#     )

#     return data


# def frequency_data(data, frequency, bandwidth, dims=None, coords=None):
#     if not isinstance(data, xr.DataArray):
#         if dims is None:
#             if data.ndim == 1:
#                 dims = 'frequency'
#             else:
#                 raise ValueError(f'Cannot guess dimensions for frequency data with {data.ndim} dimensions')
#         data = xr.DataArray(data, dims=dims)
#     data = data.assign_coords(
#         frequency=frequency,
#         bandwidth=('frequency', np.broadcast_to(bandwidth, np.shape(frequency))),
#          **{name: coord for (name, coord) in (coords or {}).items() if name != 'frequency'}
#     )
#     return data


# def time_frequency_data(data, start_time, samplerate, frequency, bandwidth, dims=None, coords=None):
#     if not isinstance(data, xr.DataArray):
#         if dims is None:
#             raise ValueError('Cannot guess dimensions for time-frequency data')
#         data = xr.DataArray(data, dims=dims)
#     data = time_data(data, start_time=start_time, samplerate=samplerate)
#     data = frequency_data(data, frequency, bandwidth)
#     return data.assign_coords(**{name: coord for (name, coord) in (coords or {}).items() if name not in ('time', 'frequency', 'bandwidth')})

class Recording(abc.ABC):
    # class _Sampling(abc.ABC):
    #     def __init__(self, recording):
    #         self.recording = recording
    #
    #     @property
    #     @abc.abstractmethod
    #     def rate(self):
    #         ...
    #
    #     @property
    #     @abc.abstractmethod
    #     def window(self):
    #         ...
    #
    #     @abc.abstractmethod
    #     def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
    #         ...

    # self.sampling = self._Sampling(self)  # pylint: disable=abstract-class-instantiated

class RecordingArray(Recording):
    # class _Sampling(Recording._Sampling):
    #     @property
    #     def rate(self):
    #         rates = [recording.sampling.rate for recording in self.recording.recordings.values()]
    #         if np.ptp(rates) == 0:
    #             return rates[0]
    #         return xr.DataArray(rates, dims='sensor', coords={'sensor': list(self.recording.recordings.keys())})

    #     @property
    #     def window(self):
    #         windows = [recording.sampling.window for recording in self.recording.recordings.values()]
    #         return positional.TimeWindow(
    #             start=max(w.start for w in windows),
    #             stop=min(w.stop for w in windows),
    #         )

    #     def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
    #         subwindow = self.window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
    #         return type(self.recording)(*[
    #             recording.sampling.subwindow(subwindow)
    #             for recording in self.recording.recordings.values()
    #         ])

    def __init__(self, *recordings):
        # self.sampling = self._Sampling(self)

class FileRecording(Recording):
        # class _Sampling(Recording._Sampling):
    #     @property
    #     def rate(self):
    #         return self.recording.files[0].samplerate

    #     @property
    #     def window(self):
    #         try:
    #             return self._window
    #         except AttributeError:
    #             self._window = positional.TimeWindow(
    #                 start=self.recording.files[0].start_time,
    #                 stop=self.recording.files[-1].stop_time,
    #             )
    #         return self._window

    #     def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
    #         original_window = self.window
    #         new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
    #         new = type(self.recording)(
    #             files=self.recording.files,
    #             sensor=self.recording.sensor,
    #         )
    #         new.sampling._window = new_window
    #         return new



import numpy as np
from . import positional


class MlogR:
    def __init__(self, m=20):
        self.m = m

    def distance_factor(self, distance, **kwargs):
        """The power reduction factor for distance.

        The factor is -M log(r), but in power-scale.
        For `m=20`, this is equivalent with 1 / r**2.
        """
        level = self.m * np.log10(distance)
        return 10 ** (-level / 10)

    def __call__(self, input_power, receiver, source, **kwargs):
        distance = positional.distance_between(source, receiver)
        return input_power / self.distance_factor(distance)


class SmoothLloydMirror(MlogR):
    def __init__(self, m=20, speed_of_sound=1500):
        super().__init__(m=m)
        self.speed_of_sound = speed_of_sound

    def surface_factor(self, grazing_angle, source_depth, frequency):
        """The factor by which a surface modifies a source power.

        For high frequencies this is a plain factor 2, since we are interested in the average value.
        For low frequencies, this is (2kd sin(θ))**2, where θ is the grazing angle.
        This is mixed as 1 / (1 / lf + 1 / hf).
        """
        kd = 2 * np.pi * frequency * source_depth / self.speed_of_sound

        mirror_lf = 4 * kd**2 * np.sin(grazing_angle)**2
        mirror_hf = 2
        mirror_reduction = 1 / (1 / mirror_lf + 1 / mirror_hf)
        return mirror_reduction

    def __call__(self, input_power, receiver, source, **kwargs):
        horizontal_distance = positional.distance_between(source, receiver)
        distance = (horizontal_distance**2 + receiver.depth**2)**0.5
        grazing_angle = np.arctan2(receiver.depth, horizontal_distance)

        # kd = 2 * np.pi * input_power.frequency * source.depth / self.speed_of_sound

        # mirror_lf = 4 * kd**2 * np.sin(np.sin(grazing_angle))**2
        # mirror_hf = 2
        # mirror_reduction = (1 / mirror_lf + 1 / mirror_hf)

        # distance_level = self.m * np.log10(distance)
        # distance_reduction = 10 ** (distance_level / 10)

        total_factor = self.distance_factor(distance) * self.surface_factor(grazing_angle=grazing_angle, source_depth=source.depth, frequency=input_power.frequency)
        return input_power / total_factor


class SeabedCriticalAngle(SmoothLloydMirror):
    def __init__(self, water_depth, m=20, speed_of_sound=1500, substrate_compressional_speed=1500):
        super().__init__(m=m, speed_of_sound=speed_of_sound)
        self.substrate_compressional_speed = substrate_compressional_speed
        self.water_depth = water_depth

    def bottom_factor(self, frequency, source_depth, distance):
        """The factor by which a bottom contains the power from a source.

        The general idea is that power radiated towards the bottom will either
        stay in the water column, and thus arrive at the receiver at some point,
        or get transmitted into the substrate.
        The grazing angle below which power will be contained is the critical angle ψ.
        For high frequencies, the bottom retains 2ψ of the energy, on average.
        For low frequencies, we have a retention of 2 (kd)**2 (ψ - sin(ψ) cos(ψ)).
        This is then distance propagated by 1/(rH) instead of 1/r**2, to accommodate the cylindrical domain.
        The low-high frequency mixing is then done as 1 / (1 / lf + 1 / hf), as for smooth Lloyd mirror.

        Note that the output from this function is compensated for the spherical distance propagation.
        It is thus expected to additionally multiply by 1 / r**2, i.e. the output here ~ r/H.
        """
        critical_angle = np.arccos(self.speed_of_sound / self.substrate_compressional_speed)
        kd = 2 * np.pi * frequency * source_depth / self.speed_of_sound
        lf_approx = 2 * kd**2 * (critical_angle - np.sin(critical_angle) * np.cos(critical_angle))
        hf_approx = 2 * critical_angle
        bottom_approx = 1 / (1 / lf_approx + 1 / hf_approx) / self.water_depth * distance  # We multiply with the distance here since we will add a MlogR as well.
        return bottom_approx

    def __call__(self, input_power, receiver, source, **kwargs):
        horizontal_distance = positional.distance_between(source, receiver)
        distance = (horizontal_distance**2 + receiver.depth**2)**0.5
        grazing_angle = np.arctan2(receiver.depth, horizontal_distance)

        distance_factor = self.distance_factor(distance=distance)
        direct_surface_factor = self.surface_factor(grazing_angle=grazing_angle, source_depth=source.depth, frequency=input_power.frequency)
        bottom_factor = self.bottom_factor(frequency=input_power.frequency, source_depth=source.depth, distance=distance)

        total_factor = distance_factor * (direct_surface_factor + bottom_factor)
        return input_power / total_factor


"""Seabed properties.

Properties included are grain size and speed of sound (compressional).
Based on Ainslie, M.A. Principles of Sonar Performance Modeling, Springer-Verlag Berlin Heidelberg, 2010.
"""
seabed_properties = {
    'very coarse sand': {
        'grain size': -0.5,
        'speed of sound': 1500 * 1.307,
    },
    'coarse sand': {
        'grain size': 0.5,
        'speed of sound': 1500 * 1.250,
    },
    'medium sand': {
        'grain size': 1.5,
        'speed of sound': 1500 * 1.198,
    },
    'fine sand': {
        'grain size': 2.5,
        'speed of sound': 1500 * 1.152,
    },
    'very fine sand': {
        'grain size': 3.5,
        'speed of sound': 1500 * 1.112,
    },
    'coarse silt': {
        'grain size': 4.5,
        'speed of sound': 1500 * 1.077,
    },
    'medium silt': {
        'grain size': 5.5,
        'speed of sound': 1500 * 1.048,
    },
    'fine silt': {
        'grain size': 6.5,
        'speed of sound': 1500 * 1.024,
    },
    'very fine silt': {
        'grain size': 7.5,
        'speed of sound': 1500 * 1.005,
    },
}



from setuptools import setup
import sys
import os.path

package_name = 'uwacan'
sys.path.insert(0, os.path.join(os.path.dirname(__file__), package_name))
from _version import version_manager  # We cannot import the _version module, but we can import from it.

with version_manager() as version:
    setup(
        name=package_name,
        version=version,
        description='Underwater Acoustic Analysis tools',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        url='https://github.com/CarlAndersson/underwater-acoustics-analysis',
        author='Carl Andersson',
        author_email='carl.andersson@ivl.se',
        packages=[package_name],
        python_requires=">=3.8",
        install_requires=[
            'numpy',
            'scipy',
            'geographiclib',
            'pendulum>=3',
            'xarray',
            'python-dotenv',
            'pysoundfile',
        ],
    )



import numpy as np
import scipy.signal
from ._core import TimePeriod
import abc


class Data:
    def __init__(self, data, dims):
        self.data = data
        self.dims = dims

    def _dims_to_axes(self, dim):
        if isinstance(dim, (int, str)):
            dim = [dim]
        included_axes = []
        for ax in dim:
            if isinstance(ax, str):
                ax = self.dims.index(ax)
            included_axes.append(ax)
        included_axes = tuple(included_axes)
        excluded_dims = tuple(dim for idx, dim in enumerate(self.dims) if idx not in included_axes)
        return included_axes, excluded_dims

    def reduce(self, function, dim, *args, **kwargs):
        reduce_axes, new_dims = self._dims_to_axes(dim)
        data = function(self.data, axis=reduce_axes, *args, **kwargs)
        
        if not isinstance(data, Data):
            try:
                data = type(self)(data=data, dims=new_dims)
            except TypeError:
                data = Data(data=data, dims=new_dims)
        return data


class TimeData(Data):
    @classmethod
    def calibrated(cls, data, calibration, **kwargs):
        data = np.asarray(data)
        calibration = np.asarray(calibration).astype('float32')
        c = 10**(calibration / 20) / 1e-6  # Calibration values are given as dB re. 1μPa
        c = c.reshape((-1,) + (1,) * (data.ndim - 1))
        if data.dtype in (np.int8, np.int16, np.int32, np.float32):
            c = c.astype(np.float32)

        calibrated = data / c
        obj = cls(data=calibrated, **kwargs)
        return obj

    def __init__(self, data, start_time, samplerate, dims=None, *args, **kwargs):
        self._start_time = start_time
        self.samplerate = samplerate
        
        if dims is None:
            if data.ndim == 1:
                dims = 'time',
            elif data.ndim == 2:
                dims = 'channels', 'time'

        super().__init__(data=data, dims=dims, *args, **kwargs)

    @property
    def time_period(self):
        return TimePeriod(start=self._start_time, duration=self.data.shape[-1] / self.samplerate)

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        window = self.time_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        start = (window.start - self._start_time).total_seconds()
        stop = (window.stop - self._start_time).total_seconds()
        # Indices assumed to be seconds from start
        start = np.math.floor(start * self.samplerate)
        stop = np.math.ceil(stop * self.samplerate)

        return type(self)(
            data=self.data[..., start:stop],
            start_time=self._start_time.add(seconds=start / self.samplerate),
            samplerate=self.samplerate,
            dims=self.dims,
        )

    def reduce(self, function, dim, *args, **kwargs):
        reduce_axes, new_dims = self._dims_to_axes(dim)
        data = function(self.data, axis=reduce_axes, *args, **kwargs)
        
        if not isinstance(data, Data):
            if 'time' in new_dims:
                data = TimeData(data, start_time=self.start_time, samplerate=self.samplerate, dims=new_dims)
            else:
                data = Data(data, dims=new_dims)
        return data


class FrequencyData(Data):
    def __init__(self, data, frequency, bandwidth, dims=None, *args, **kwargs):
        self.frequency = frequency
        self.bandwidth = bandwidth

        if dims is None:
            if data.ndim == 1:
                dims = 'frequency',
            elif data.ndim == 2:
                dims = 'channels', 'frequency'

        super().__init__(data=data, dims=dims, *args, **kwargs)

    def reduce(self, function, dim, *args, **kwargs):
        reduce_axes, new_dims = self._dims_to_axes(dim)
        data = function(self.data, axis=reduce_axes, *args, **kwargs)

        if not isinstance(data, Data):
            if 'frequency' in new_dims:
                data = FrequencyData(data, frequency=self.frequency, bandwidth=self.bandwidth, dims=new_dims)
            else:
                data = Data(data, dims=new_dims)
        return data


class TimeFrequencyData(Data):
    def __init__(self, data, start_time, samplerate, frequency, bandwidth, dims=None, *args, **kwargs):
        self.frequency = frequency
        self.bandwidth = bandwidth
        self._start_time = start_time
        self.samplerate = samplerate

        if dims is None:
            if data.ndim == 2:
                dims = 'frequency', 'time'
            elif data.ndim == 3:
                dims = 'channels', 'frequency', 'time'

        super().__init__(data=data, dims=dims, *args, **kwargs)

    @property
    def time_period(self):
        return TimePeriod(start=self._start_time, duration=self.data.shape[-1] / self.samplerate)

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        window = self.time_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        start = (window.start - self._start_time).total_seconds()
        stop = (window.stop - self._start_time).total_seconds()
        # Indices assumed to be seconds from start
        start = np.math.floor(start * self.samplerate)
        stop = np.math.ceil(stop * self.samplerate)
        return type(self)(
            data=self.data[..., start:stop],
            start_time=self._start_time.add(seconds=start / self.samplerate),
            samplerate=self.samplerate,
            frequency=self.frequency,
            bandwidth=self.bandwidth,
            dims=self.dims,
        )

    def reduce(self, function, dim, *args, **kwargs):
        reduce_axes, new_dims = self._dims_to_axes(dim)
        data = function(self.data, axis=reduce_axes, *args, **kwargs)

        if not isinstance(data, Data):
            if 'frequency' in new_dims and 'time' in new_dims:
                data = TimeFrequencyData(
                    data,
                    start_time=self.start_time,
                    samplerate=self.samplerate,
                    frequency=self.frequency,
                    bandwidth=self.bandwidth,
                    dims=new_dims,
                )
            elif 'frequency' in new_dims:
                data = FrequencyData(data, frequency=self.frequency, bandwidth=self.bandwidth, dims=new_dims)
            elif 'time' in new_dims:
                data = TimeData(data, start_time=self.start_time, samplerate=self.samplerate, dims=new_dims)
            else:
                data = Data(data, dims=new_dims)
        return data


class Spectrogram(TimeFrequencyData):
    @classmethod
    def from_time_data(cls, time_data, window_duration, window='hann', overlap=0.5):
        if not isinstance(time_data, TimeData):
            raise TypeError(f"Cannot calculate the spectrogram of object of type '{time_data.__class__.__name__}'")
        window_samples = round(window_duration * time_data.samplerate)
        overlap_samples = round(window_duration * overlap * time_data.samplerate)
        f, t, Sxx = scipy.signal.spectrogram(
            x=time_data.data,
            fs=time_data.samplerate,
            window=window,
            nperseg=window_samples,
            noverlap=overlap_samples,
        )
        return cls(
            data=Sxx.copy(),  # Using a copy here is a performance improvement in later processing stages.
            # The array returned from the spectrogram function is the real part of the original stft, reshaped.
            # This means that the array takes twice the memory (the imaginary part is still around),
            # and it's not contiguous which slows down filtering a lot.
            samplerate=time_data.samplerate / (window_samples - overlap_samples),
            start_time=time_data.time_period.start.add(seconds=t[0]),
            frequency=f,
            bandwidth=time_data.samplerate / window_samples,
        )
